"""
Utility functions for reading MITgcm mds files (.meta / .data)
"""
# python 3 compatiblity
from __future__ import print_function, division

import re
import os
import numpy as np
import warnings
from functools import reduce
from dask import delayed
import dask.array as dsa
from dask.base import tokenize
import xarray as xr
import sys

def parse_meta_file(fname):
    """Get the metadata as a dict out of the MITgcm mds .meta file.

    PARAMETERS
    ----------
    fname : str
        Path to the .meta file

    RETURNS
    -------
    flds : dict
        Metadata in dictionary form.
    """
    flds = {}
    basename = re.match('(^.+?)\..+', os.path.basename(fname)).groups()[0]
    flds['basename'] = basename
    with open(fname) as f:
        text = f.read()
    # split into items
    for item in re.split(';', text):
        # remove whitespace at beginning
        item = re.sub('^\s+', '', item)
        match = re.match('(\w+) = (\[|\{)(.*)(\]|\})', item, re.DOTALL)
        if match:
            key, _, value, _ = match.groups()
            # remove more whitespace
            value = re.sub('^\s+', '', value)
            value = re.sub('\s+$', '', value)
            # print key,':', value
            flds[key] = value
    # now check the needed things are there
    needed_keys = ['dimList', 'nDims', 'nrecords', 'dataprec']
    for k in needed_keys:
        assert k in flds
    # transform datatypes
    flds['nDims'] = int(flds['nDims'])
    flds['nrecords'] = int(flds['nrecords'])
    # endianness is set by _read_mds
    flds['dataprec'] = np.dtype(re.sub("'", '', flds['dataprec']))
    flds['dimList'] = [[int(h) for h in
                       re.split(',', g)] for g in
                       re.split(',\n', flds['dimList'])]
    if 'fldList' in flds:
        flds['fldList'] = [re.match("'*(\w+)", g).groups()[0] for g in
                           re.split("'\s+'", flds['fldList'])]
        assert flds['nrecords'] == len(flds['fldList'])
    return flds

def _get_useful_info_from_meta_file(metafile):
    # why does the .meta file contain so much repeated info?
    # Here we just get the part we need
    # and reverse order (numpy uses C order, mds is fortran)
    meta = parse_meta_file(metafile)
    shape = [g[0] for g in meta['dimList']][::-1]
    assert len(shape) == meta['nDims']
    # now add an extra for number of recs
    nrecs = meta['nrecords']
    shape.insert(0, nrecs)
    dtype = meta['dataprec']
    if 'fldList' in meta:
        fldlist = meta['fldList']
        name = fldlist[0]
    else:
        name = meta['basename']
        fldlist = None

    return nrecs, shape, name, dtype, fldlist


def read_mds(fname, iternum=None, use_mmap=None, endian='>', shape=None,
             dtype=None, use_dask=True, extra_metadata=None, chunks="3D",
             llc=False, llc_method="smallchunks", legacy=True):
    """Read an MITgcm .meta / .data file pair


    PARAMETERS
    ----------
    fname : str
        The base name of the data file pair (without a .data or .meta suffix)
    iternum : int, optional
        The iteration number suffix
    use_mmap : bool, optional
        Whether to read the data using a numpy.memmap.
        Mutually exclusive with `use_dask`.
    endian : {'>', '<', '|'}, optional
        Dndianness of the data
    dtype : numpy.dtype, optional
        Data type of the data (will be inferred from the .meta file by default)
    shape : tuple, optional
        Shape of the data (will be inferred from the .meta file by default)
    use_dask : bool, optional
        Whether wrap the reading of the raw data in a ``dask.delayed`` object.
        Mutually exclusive with `use_mmap`.
    extra_metadata : dict, optional
        Dictionary containing some extra metadata that will be appended to
        content of MITgcm meta file to create the file_metadata. This is needed
        for llc type configurations (global or regional). In this case the
        extra metadata used is of the form :

        aste = {'has_faces': True, 'ny': 1350, 'nx': 270,
        'ny_facets': [450,0,270,180,450],
        'pad_before_y': [90,0,0,0,0],
        'pad_after_y': [0,0,0,90,90],
        'face_facets': [0, 0, 2, 3, 4, 4],
        'facet_orders' : ['C', 'C', 'C', 'F', 'F'],
        'face_offsets' : [0, 1, 0, 0, 0, 1],
        'transpose_face' : [False, False, False,
        True, True, True]}

        llc90 = {'has_faces': True, 'ny': 13*90, 'nx': 90,
        'ny_facets': [3*90, 3*90, 90, 3*90, 3*90],
        'face_facets': [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4],
        'facet_orders': ['C', 'C', 'C', 'F', 'F'],
        'face_offsets': [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2],
        'transpose_face' : [False, False, False,
        False, False, False, False,
        True, True, True, True, True, True]}

        llc grids have typically 5 rectangular facets and will be mapped onto
        N (=13 for llc, =6 for aste) square faces.
        Keys for the extra_metadata dictionary can be of different types and
        length:


        * bool:

        #. has_faces : True if domain is combination of connected grids

        * list of len=nfacets:

        #. ny_facets : number of points in y direction of each facet
        (usually n * nx)
        #. pad_before_y (Regional configuration) : pad data with N zeros
        before array
        #. pad_after_y (Regional configuration) : pad data with N zeros
        after array
        #. facet_order : row/column major order of this facet

        * list of len=nfaces:

        #. face_facets : facet of origin for this face

        #. face_offsets : position of the face in the facet (0 = start)

        #. transpose_face : transpose the data for this face

    chunks : {'3D', '2D'}
        Which routine to use for chunking data. '2D' splits the file
        into a individual dask chunk of size (nx x nx) for each face (if llc)
        of each record of each level.
        '3D' loads the whole raw data file (either into memory or as a
        numpy.memmap) and is not suitable for llc configurations.
        The different methods will have different memory and i/o performance
        depending on the details of the system configuration.

    obsolete : llc and llc_methods, kept for testing

    RETURNS
    -------
    data : dict
       The keys correspond to the variable names of the different variables in
       the data file. The values are the data itself, either as an
       ``numpy.ndarray``, ``numpy.memmap``, or ``dask.array.Array`` depending
       on the options selected.
    """

    if use_mmap and use_dask:
        raise TypeError('`use_mmap` and `use_dask` are mutually exclusive:'
                        ' Both memory-mapped and dask arrays'
                        ' use lazy evaluation.')
    elif use_mmap is None:
        use_mmap = False if use_dask else True

    if iternum is None:
        istr = ''
    else:
        assert isinstance(iternum, int)
        istr = '.%010d' % iternum
    datafile = fname + istr + '.data'
    metafile = fname + istr + '.meta'

    if use_mmap and use_dask:
        raise TypeError('nope')
    elif use_mmap is None:
        use_mmap = False if use_dask else True

    # get metadata
    try:
        metadata = parse_meta_file(metafile)
        nrecs, shape, name, dtype, fldlist = \
            _get_useful_info_from_meta_file(metafile)
        dtype = dtype.newbyteorder(endian)
    except IOError:
        # we can recover from not having a .meta file if dtype and shape have
        # been specified already
        if shape is None:
            raise IOError("Cannot find the shape associated to %s in the \
                          metadata." % fname)
        elif dtype is None:
            raise IOError("Cannot find the dtype associated to %s in the \
                          metadata, please specify the default dtype to \
                          avoid this error." % fname)
        else:
            # add time dimensions
            shape = (1,) + shape
            shape = list(shape)
            name = os.path.basename(fname)

            metadata = {'basename': name, 'shape': shape}

    # figure out dimensions
    ndims = len(shape)-1
    if ndims == 3:
        _, nz, ny, nx = shape
        dims_vars = ('nz', 'ny', 'nx')
    elif ndims == 2:
        _, ny, nx = shape
        nz = 1
        dims_vars = ('ny', 'nx')

    # and variables
    if 'fldList' not in metadata:
        metadata['fldList'] = [metadata['basename']]

    # if not provided in extra_metadata, we assume that the variables in file
    # have the same shape
    if extra_metadata is None or 'dims_vars' not in extra_metadata:
        dims_vars_list = []
        for var in metadata['fldList']:
            dims_vars_list.append(dims_vars)

    # add extra dim information and set aside
    metadata.update({'dims_vars': dims_vars_list,
                     'dtype': dtype, 'endian': endian,
                     'nx': nx, 'ny': ny,
                     'nz': nz, 'nt': 1})  # parse_meta harcoded for nt = 1

    file_metadata = metadata.copy()

    # by default, we set to non-llc grid
    file_metadata.update({'filename': datafile, 'vars': metadata['fldList'],
                          'has_faces': False})

    # extra_metadata contains informations about llc/regional llc grid
    if extra_metadata is not None:
        nhpts_ex = extra_metadata['nx'] * extra_metadata['ny']
        nhpts = metadata['nx'] * metadata['ny']
        # check that nx * ny is consistent between extra_metadata and meta file
        # unless it's a vertical profile nx = ny = 1
        if nhpts > 1:
            assert nhpts_ex == nhpts
        file_metadata.update(extra_metadata)

    # --------------- LEGACY --------------------------
    # from legacy code (needs to be phased out)
    # transition code to keep unit tests working
    if llc:
        chunks = "2D"
    # --------------- /LEGACY --------------------------

    # it is possible to override the values of nx, ny, nz from extra_metadata
    # (needed for bug meta file ASTE) except if those are = 1 (vertical coord)
    # where we override by values found in meta file
    for dim in ['nx', 'ny', 'nz']:
        if metadata[dim] == 1:
            file_metadata.update({dim: 1})

    # read all variables from file into the list d
    d = read_all_variables(file_metadata['fldList'], file_metadata,
                           use_mmap=use_mmap, use_dask=use_dask,
                           chunks=chunks)

    # convert list into dictionary
    out = {}
    for n, name in enumerate(file_metadata['fldList']):
        if ndims == 3:
            out[name] = d[n]
        elif ndims == 2:
            out[name] = d[n][:, 0, :]

    # --------------- LEGACY --------------------------
    # from legacy code (needs to be phased out)
    # transition code to keep unit tests working
    if legacy:
        for n, name in enumerate(file_metadata['fldList']):
            out[name] = out[name][0, :]
    # --------------- /LEGACY --------------------------
    return out


def read_raw_data(datafile, dtype, shape, use_mmap=False, offset=0,
                  order='C', partial_read=False):
    """Read a raw binary file and shape it.

    PARAMETERS
    ----------
    datafile : str
        Path to a .data file
    dtype : numpy.dtype
        Data type of the data
    shape : tuple
        Shape of the data
    use_memmap : bool, optional
        Whether to read the data using a numpy.memmap
    offset : int, optional
        Offset (in bytes) to apply on read
    order : str, optional
        Row/Column Major = 'C' or 'F'
    partial_read : bool, optional
        If reading part of the file

    RETURNS
    -------
    data : numpy.ndarray
        The data (or a memmap to it)
    """

    number_of_values = reduce(lambda x, y: x * y, shape)
    expected_number_of_bytes = number_of_values * dtype.itemsize
    actual_number_of_bytes = os.path.getsize(datafile)
    if not partial_read:
        # first check that partial_read and offset are used together
        if offset != 0:
            raise ValueError(
                'When partial_read==False, offset will not be read')
        # second check to be sure there is the right number of bytes in file
        if expected_number_of_bytes != actual_number_of_bytes:
            raise IOError('File `%s` does not have the correct size '
                          '(expected %g, found %g)' %
                          (datafile,
                           expected_number_of_bytes,
                           actual_number_of_bytes))
    else:
        pass

    if offset < actual_number_of_bytes:
        pass
    else:
        raise ValueError('bytes offset %g is greater than file size %g' %
                         (offset, actual_number_of_bytes))

    with open(datafile, 'rb') as f:
        if use_mmap:
            data = np.memmap(f, dtype=dtype, mode='r', offset=offset,
                             shape=tuple(shape), order=order)
        else:
            f.seek(offset)
            data = np.fromfile(f, dtype=dtype, count=number_of_values)
            data = data.reshape(shape, order=order)
    data.shape = shape
    return data


def parse_namelist(file, silence_errors=True):
    """Read a FOTRAN namelist file into a dictionary.

    PARAMETERS
    ----------
    file : str
        Path to the namelist file to read.

    RETURNS
    -------
    data : dict
        Dictionary of each namelist as dictionaries
    """
    def parse_val(val):
        """Parse a string and cast it in the appropriate python type."""
        if ',' in val:  # It's a list, parse recursively
            return [parse_val(subval.strip()) for subval in val.split(',')]
        elif val.startswith("'"):  # It's a string, remove quotes.
            return val[1:-1].strip()
        elif '*' in val:  # It's shorthand for a repeated value
            repeat, number = val.split('*')
            return [parse_val(number)] * int(repeat)
        elif val in ['.TRUE.', '.FALSE.']:
            return val == '.TRUE.'
        elif '.' in val or 'E' in val:  # It is a Real (float)
            return float(val)
        # Finally try for an int
        return int(val)

    data = {}
    current_namelist = ''
    raw_lines = []
    with open(file) as f:
        for line in f:
            # Remove comments
            line = line.split('#')[0].strip()
            if '=' in line or '&' in line:
                raw_lines.append(line)
            elif line:
                raw_lines[-1] += line

    for line in raw_lines:
        if line.startswith('&'):
            current_namelist = line.split('&')[1]
            if current_namelist:  # else : it's the end of a namelist.
                data[current_namelist] = {}
        else:
            field, value = map(str.strip, line[:-1].split('='))
            try:
                value = parse_val(value)
            except ValueError:
                mess = ('Unable to read value for field {field} in file {file}: {value}'
                        ).format(field=field, file=file, value=value)
                if silence_errors:
                    warnings.warn(mess)
                    value = None
                else:
                    raise ValueError(mess)

            if '(' in field:  # Field is an array
                field, idxs = field[:-1].split('(')
                if field not in data[current_namelist]:
                    data[current_namelist][field] = []
                # For generality, we will assign a slice, so we cast in list
                value = value if isinstance(value, list) else [value]
                idxs = [slice(int(idx.split(':')[0]) - 1,
                              int(idx.split(':')[1]))
                        if ':' in idx else slice(int(idx) - 1, int(idx))
                        for idx in idxs.split(',')]

                datafield = data[current_namelist][field]
                # Array are 1D or 2D, if 2D we extend it to the good shape,
                # filling it with [] and pass the appropriate sublist.
                # Only works with slice assign (a:b) in first position.
                missing_spots = idxs[-1].stop - len(datafield)
                if missing_spots > 0:
                    datafield.extend([] for i in range(missing_spots))
                if len(idxs) == 2:
                    datafield = datafield[idxs[1].start]
                datafield[idxs[0]] = value
            else:
                data[current_namelist][field] = value
    return data


def parse_available_diagnostics(fname, layers={}):
    """Examine the available_diagnostics.log file and translate it into
    useful variable metadata.

    PARAMETERS
    ----------
    fname : str or buffer
        the path to the diagnostics file or a file buffer
    layers : dict (optional)
        dictionary mapping layers names to dimension sizes

    RETURNS
    -------
    all_diags : a dictionary keyed by variable names with values
        (coords, description, units)
    """
    all_diags = {}
    diag_id_lookup = {}
    mate_lookup = {}

    # mapping between the available_diagnostics.log codes and the actual
    # coordinate names
    # http://mitgcm.org/public/r2_manual/latest/online_documents/node268.html
    xcoords = {'U': 'i_g', 'V': 'i', 'M': 'i', 'Z': 'i_g'}
    ycoords = {'U': 'j', 'V': 'j_g', 'M': 'j', 'Z': 'j_g'}
    rcoords = {'M': 'k', 'U': 'k_u', 'L': 'k_l'}

    # need to be able to accept string filename or buffer
    def process_buffer(f):
        for l in f:
            # will automatically skip first four header lines
            c = re.split('\|', l)
            if len(c) == 7 and c[0].strip() != 'Num':
                # parse the line to extract the relevant variables
                key = c[1].strip()
                diag_id = int(c[0].strip())
                diag_id_lookup[diag_id] = key
                levs = int(c[2].strip())
                mate = c[3].strip()
                if mate:
                    mate = int(mate)
                    mate_lookup[key] = mate
                code = c[4]
                units = c[5].strip()
                desc = c[6].strip()

                # decode what those variables mean
                hpoint = code[1]
                rpoint = code[8]
                xycoords = [ycoords[hpoint], xcoords[hpoint]]
                rlev = code[9]

                if rlev == '1' and levs == 1:
                    zcoord = []
                elif rlev == 'R':
                    zcoord = [rcoords[rpoint]]
                elif rlev == 'L':  # pragma : no cover
                    # max(Nr, Nrphys) according to doc...
                    # this seems to be only used in atmos
                    # with different levels for dynamics and physics
                    # setting to Nr meanwhile
                    zcoord = [rcoords[rpoint]]
                elif rlev == 'X':
                    if layers:
                        layer_name = key.ljust(8)[-4:].strip()
                        n_layers = layers[layer_name]
                        if levs == n_layers:
                            suffix = 'bounds'
                        elif levs == (n_layers-1):
                            suffix = 'center'
                        elif levs == (n_layers-2):
                            suffix = 'interface'
                        else:  # pragma: no cover
                            suffix = None
                            warnings.warn("Could not match rlev = %g to a "
                                          "layers coordiante" % rlev)
                        # dimname = ('layer_' + layer_name + '_' +
                        #            suffix if suffix
                        dimname = (('l' + layer_name[0] + '_' +
                                    suffix[0]) if suffix else '_UNKNOWN_')
                        zcoord = [dimname]
                    else:
                        zcoord = ['_UNKNOWN_']
                else:  # pragma: no cover
                    warnings.warn("Not sure what to do with rlev = " + rlev)
                    warnings.warn("corresponding diag_id  = " + str(diag_id))
                    zcoord = ['_UNKNOWN_']
                coords = zcoord + xycoords
                all_diags[key] = dict(dims=coords,
                                      # we need a standard name
                                      attrs={'standard_name': key,
                                             'long_name': desc,
                                             'units': units})
    try:
        with open(fname) as f:
            process_buffer(f)
    except TypeError:
        process_buffer(fname)

    # add mate information
    for key, mate_id in mate_lookup.items():
        all_diags[key]['attrs']['mate'] = diag_id_lookup[mate_id]
    return all_diags

# stuff related to llc mds file structure
LLC_NUM_FACES=13
facet_strides = ((0,3), (3,6), (6,7), (7,10), (10,13))
facet_orders = ('C', 'C', 'C', 'F', 'F')
face_facets = [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4]
face_offsets = [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2]
transpose_face = [False, False, False, False, False, False, False,
                  True, True, True, True, True, True]

def _read_2d_facet(fname, nfacet, nlev, nx, dtype='>f8', memmap=True):
    # make sure we have a valid dtype
    dtype = np.dtype(dtype)
    nbytes = dtype.itemsize

    # where the facet starts in the file
    facet_offset = facet_strides[nfacet][0] * nx * nx * nbytes
    level_offset = LLC_NUM_FACES * nx * nx * nbytes * nlev
    offset = facet_offset + level_offset

    # the array order of the facet
    facet_order = facet_orders[nfacet]

    # the size shape of the facet
    facet_ny = (facet_strides[nfacet][1] - facet_strides[nfacet][0])*nx
    facet_shape = (facet_ny, nx)
    facet_nitems = facet_ny * nx
    with open(fname, 'rb') as f:
        #print("Reading %s facet %g nlev %g" % (fname, nfacet, nlev))
        if memmap:
            data = np.memmap(f, dtype=dtype, mode='r', offset=offset,
                             shape=facet_shape, order=facet_order)
        else:
            f.seek(offset)
            data = np.fromfile(f, dtype=dtype, count=facet_nitems)
            data = data.reshape(facet_shape, order=facet_order)
    return data

def _read_2d_face(fname, nface, nlev, nx, dtype='>f8', memmap=True):
    # make sure we have a valid dtype
    nfacet = face_facets[nface]
    face_slice = slice(nx*face_offsets[nface], nx*(face_offsets[nface]+1))
    facet_offset = nx * face_offsets[nface]
    data_facet = _read_2d_facet(fname, nfacet, nlev, nx,
                               dtype=dtype, memmap=memmap)
    data = data_facet[face_slice]
    if transpose_face[nface]:
        data = data.T
    return data

# manually construct dask graph
def read_3d_llc_data(fname, nz, nx, dtype='>f8', memmap=True, nrecs=1,
                     method="smallchunks"):
    """Read a three-dimensional LLC file using a custom dask graph.

    PARAMETERS
    ----------
    fname : string
        Path to the file on disk
    nz : int
        Number of vertical levels
    nx : int
        Size of each face side dimension
    dtype : np.dtype, optional
        Datatype of the data
    memmap : bool, optional
        Whether to read the data using np.memmap. Forced to be ``False`` for
        ``method="smallchunks"``.
    nrecs : int, optional
        The number of records in a multi-record file
    method : {"smallchunks", "bigchunks"}, optional
        Which routine to use for reading raw LLC. "smallchunks" splits the file
        into a individual dask chunk of size (nx x nx) for each face of each
        level (i.e. the total number of chunks is 13 * nz). "bigchunks" loads
        the whole raw data file (either into memory or as a numpy.memmap),
        splits it into faces, and concatenates those faces together using
        ``dask.array.concatenate``. The different methods will have different
        memory and i/o performance depending on the details of the system
        configuration.

    RETURNS
    -------
    data : dask.array.Array
        The data
    """
    dtype = np.dtype(dtype)

    if method == "smallchunks":

        def load_chunk(nface, nlev):
            return _read_2d_face(fname, nface, nlev, nx,
                                 dtype=dtype, memmap=memmap)[None, None, None]

        chunks = (1, 1, 1, nx, nx)
        shape = (nrecs, nz, LLC_NUM_FACES, nx, nx)
        name = 'llc-' + tokenize(fname, shape, dtype,
                                 method)  # unique identifier
        # we hack the record number as extra vertical levels
        dsk = {(name, nrec, nlev, nface, 0, 0): (load_chunk, nface,
                                                 nlev + nz*nrec)
               for nface in range(LLC_NUM_FACES)
               for nlev in range(nz)
               for nrec in range(nrecs)}

        data = dsa.Array(dsk, name, chunks, dtype=dtype, shape=shape)

    elif method == "bigchunks":
        shape = (nrecs, nz, LLC_NUM_FACES*nx, nx)
        # the dimension that needs to be reshaped
        jdim = 2
        data = read_raw_data(fname, dtype, shape, use_mmap=memmap)
        data = _reshape_llc_data(data, jdim)

    # automatically squeeze off z dimension; this matches mds file behavior
    if nz == 1:
        data = data[:, 0]
    return data


# a deprecated function that I can't bear to delete because it was painful to
# write
def _reshape_llc_data(data, jdim):  # pragma: no cover
    """Fix the weird problem with llc data array order."""
    # Can we do this without copying any data?
    # If not, we need to go upstream and implement this at the MDS level
    # Or can we fudge it with dask?
    # this is all very specific to the llc file output
    # would be nice to generalize more, but how?
    nside = data.shape[jdim] // LLC_NUM_FACES
    # how the LLC data is laid out along the j dimension
    strides = ((0,3), (3,6), (6,7), (7,10), (10,13))
    # whether to reshape each face
    reshape = (False, False, False, True, True)
    # this will slice the data into 5 facets
    slices = [jdim * (slice(None),) + (slice(nside*st[0], nside*st[1]),)
              for st in strides]
    facet_arrays = [data[sl] for sl in slices]
    face_arrays = []
    for ar, rs, st in zip(facet_arrays, reshape, strides):
        nfaces_in_facet = st[1] - st[0]
        shape = list(ar.shape)
        if rs:
            # we assume the other horizontal dimension is immediately after jdim
            shape[jdim] = ar.shape[jdim+1]
            shape[jdim+1] = ar.shape[jdim]
        # insert a length-1 dimension along which to concatenate
        shape.insert(jdim, 1)
        # this modify the array shape in place, with no copies allowed
        # but it doesn't work with dask arrays
        # ar.shape = shape
        ar = ar.reshape(shape)
        # now ar is propery shaped, but we still need to slice it into faces
        face_slice_dim = jdim + 1 + rs
        for n in range(nfaces_in_facet):
            face_slice = (face_slice_dim * (slice(None),) +
                          (slice(nside*n, nside*(n+1)),))
            data_face = ar[face_slice]
            face_arrays.append(data_face)

    # We can't concatenate using numpy (hcat etc.) because it makes a copy,
    # presumably loading the memmaps into memory.
    # Using dask gets around this.
    # But what if we want different chunks, or already chunked the data
    # upstream? Doesn't seem like this is ideal
    # TODO: Refactor handling of dask arrays and chunking
    #return np.concatenate(face_arrays, axis=jdim)
    # the dask version doesn't work because of this:
    # https://github.com/dask/dask/issues/1645
    face_arrays_dask = [dsa.from_array(fa, chunks=fa.shape)
                        for fa in face_arrays]
    concat = dsa.concatenate(face_arrays_dask, axis=jdim)
    return concat


def _llc_face_shape(llc_id):
    """Given an integer identifier for the llc grid, return the face shape."""

    # known valid LLC configurations
    if llc_id in (90, 270, 1080, 2160, 4320):
        return (llc_id, llc_id)
    else:
        raise ValueError("%g is not a valid llc identifier" % llc_id)

def _llc_data_shape(llc_id, nz=None):
    """Given an integer identifier for the llc grid, and possibly a number of
    vertical grid points, return the expected shape of the full data field."""

    # this is a constant for all LLC setups
    NUM_FACES = 13

    face_shape = _llc_face_shape(llc_id)
    data_shape = (NUM_FACES,) + face_shape
    if nz is not None:
        data_shape = (nz,) + data_shape

    # should we accomodate multiple records?
    # no, not in this function
    return data_shape


def read_all_variables(variable_list, file_metadata, use_mmap=False,
                       use_dask=False, chunks="3D"):
    """
    Return a dictionary of dask arrays for variables in a MDS file

    PARAMETERS
    ----------
    variable_list : list
        list of MITgcm variables, from fldList in .meta
    file_metadata : dict
        internal metadata for binary file
    use_mmap : bool, optional
        Whether to read the data using a numpy.memmap
    chunks : str, optional
        Whether to read 2D (default) or 3D chunks
        2D chunks are reading (x,y) levels and 3D chunks
        are reading the a (x,y,z) field

    RETURNS
    -------
    out : list
        list of data arrays (dask.array, numpy.ndarray or memmap)
        corresponding to variables from given list in the file
        described by file_metadata

    """

    out = []
    for variable in variable_list:
        if chunks == "2D":
            out.append(read_2D_chunks(variable, file_metadata,
                                      use_mmap=use_mmap, use_dask=use_dask))
        elif chunks == "3D":
            out.append(read_3D_chunks(variable, file_metadata,
                                      use_mmap=use_mmap, use_dask=use_dask))

    return out


def read_2D_chunks(variable, file_metadata, use_mmap=False, use_dask=False):
    """
    Return dask array for variable, from the file described by file_metadata,
    reading 2D chunks.

    Parameters
    ----------
    variable : string
               name of the variable to read
    file_metadata : dict
               internal file_metadata for binary file
    use_mmap : bool, optional
               Whether to read the data using a numpy.memmap
    use_dask : bool, optional
               collect the data lazily or eagerly

    Returns
    -------
    dask array for variable, with 2d (ny, nx) chunks
    or numpy.ndarray or memmap, depending on input args

    """

    if (file_metadata['nx'] == 1) and (file_metadata['ny'] == 1) and \
       (len(file_metadata['vars']) == 1):
            # vertical coordinate
        data_raw = read_raw_data(file_metadata['filename'],
                                 file_metadata['dtype'],
                                 (file_metadata['nz'],), use_mmap=use_mmap,
                                 offset=0, order='C', partial_read=False)

        shape = (file_metadata['nt'], file_metadata['nz'], 1,
                 file_metadata['ny'], file_metadata['nx'])
        data_raw = np.reshape(data_raw, shape)  # memmap -> ndarray
        chunks = (file_metadata['nt'], 1, 1,
                  file_metadata['ny'], file_metadata['nx'])
        data = dsa.from_array(data_raw, chunks=chunks)

    else:
        if file_metadata['has_faces']:
            def load_chunk(face, lev, rec):
                return _read_xy_chunk(variable, file_metadata, rec=rec,
                                      lev=lev, face=face,
                                      use_mmap=use_mmap)[None, None, None]

            chunks = (1, 1, 1, file_metadata['nx'], file_metadata['nx'])
            shape = (file_metadata['nt'], file_metadata['nz'],
                     len(file_metadata['face_facets']),
                     file_metadata['nx'], file_metadata['nx'])
            name = 'llcmds-' + tokenize(file_metadata, variable)

            dsk = {(name, rec, lev, face, 0, 0): (load_chunk, face,
                                                  lev, rec)
                   for face in range(len(file_metadata['face_facets']))
                   for lev in range(file_metadata['nz'])
                   for rec in range(file_metadata['nt'])}

        else:
            def load_chunk(lev, rec):
                return _read_xy_chunk(variable, file_metadata,
                                      rec=rec, lev=lev,
                                      face=0, use_mmap=use_mmap)[None, None]

            chunks = (1, 1, file_metadata['ny'], file_metadata['nx'])
            shape = (file_metadata['nt'], file_metadata['nz'],
                     file_metadata['ny'], file_metadata['nx'])
            name = 'mds-' + tokenize(file_metadata, variable)

            dsk = {(name, rec, lev, 0, 0): (load_chunk, lev, rec)
                   for lev in range(file_metadata['nz'])
                   for rec in range(file_metadata['nt'])}

        data = dsa.Array(dsk, name, chunks,
                         dtype=file_metadata['dtype'], shape=shape)

    if not use_dask:
        data = data.compute()

    return data


def read_3D_chunks(variable, file_metadata, use_mmap=False, use_dask=False):
    """
    Return dask array for variable, from the file described by file_metadata,
    reading 3D chunks. Not suitable for llc data.

    Parameters
    ----------
    variable : string
               name of the variable to read
    file_metadata : dict
               internal file_metadata for binary file
    use_mmap : bool, optional
               Whether to read the data using a numpy.memmap
    use_dask : bool, optional
               collect the data lazily or eagerly

    Returns
    -------
    dask array for variable, with 3d (nz, ny, nx) chunks
    or numpy.ndarray or memmap, depending on input args

    """

    def load_chunk(rec):
        return _read_xyz_chunk(variable, file_metadata,
                               rec=rec,
                               use_mmap=use_mmap)[None]

    chunks = (1, file_metadata['nz'], file_metadata['ny'], file_metadata['nx'])
    shape = (file_metadata['nt'], file_metadata['nz'],
             file_metadata['ny'], file_metadata['nx'])
    name = 'mds-' + tokenize(file_metadata, variable)

    dsk = {(name, rec, 0, 0, 0): (load_chunk, rec)
           for rec in range(file_metadata['nt'])}

    data = dsa.Array(dsk, name, chunks,
                     dtype=file_metadata['dtype'], shape=shape)

    if not use_dask:
        data = data.compute()

    return data


def _read_xyz_chunk(variable, file_metadata, rec=0, use_mmap=False):
    """
    Read a 3d chunk (x,y,z) of variable from file described in
    file_metadata.

    Parameters
    ----------
    variable : string
               name of the variable to read
    file_metadata : dict
               file_metadata for binary file
    rec      : integer, optional
               time record to read (default=0)
    use_mmap : bool, optional
               Whether to read the data using a numpy.memmap

    Returns
    -------
    numpy array or memmap
    """

    if file_metadata['has_faces'] and ((file_metadata['nx'] > 1) or
                                       (file_metadata['ny'] > 1)):
        raise ValueError("_read_xyz_chunk cannot be called with llc type grid")

    # size of the data element
    nbytes = file_metadata['dtype'].itemsize
    # byte order
    file_metadata['datatype'] = file_metadata['dtype'].newbyteorder(
        file_metadata['endian'])
    # find index of variable
    idx_var = file_metadata['vars'].index(variable)

    # 1. compute offset_variable, init to zero
    offset_vars = 0
    # loop on variables before the one to read
    for jvar in np.arange(idx_var):
        # inspect its dimensions
        dims = file_metadata['dims_vars'][jvar]
        # compute the byte size of this variable
        nbytes_thisvar = 1*nbytes
        for dim in dims:
            nbytes_thisvar = nbytes_thisvar*file_metadata[dim]
        # update offset from previous variables
        offset_vars = offset_vars+nbytes_thisvar

    # 2. get dimensions of desired variable
    dims = file_metadata['dims_vars'][idx_var]
    # inquire for values of dimensions, else return 1
    nt, nz, ny, nx = [file_metadata.get(dimname, 1)
                      for dimname in ('nt', 'nz', 'ny', 'nx')]

    # 3. compute offset from previous records of current variable
    if (rec > nt-1):
        raise ValueError("time record %g greater than number of records %g" %
                         (rec, nt))
    else:
        offset_timerecords = rec * nz * ny * nx * nbytes

    # 4. compute the offset of the previous variables, records and levels
    offset = offset_vars + offset_timerecords
    shape = (nz, ny, nx,)

    # check if we do a partial read of the file
    if (nt > 1) or (len(file_metadata['vars']) > 1):
        partial_read = True
    else:
        partial_read = False

    # define the order (row/column major)
    # in conventional grids, it's in C
    order = 'C'

    # 5. Do the actual read
    data = read_raw_data(file_metadata['filename'],
                         file_metadata['datatype'],
                         shape, use_mmap=use_mmap, offset=offset,
                         order=order, partial_read=partial_read)

    return data


def _read_xy_chunk(variable, file_metadata, rec=0, lev=0, face=0,
                   use_mmap=False):
    """
    Read a 2d chunk along (x,y) of variable from file described in
    file_metadata.

    Parameters
    ----------
    variable : string
               name of the variable to read
    file_metadata : dict
               file_metadata for binary file
    rec      : integer, optional
               time record to read (default=0)
    lev      : integer, optional
               vertical level to read (default=0)
    face     : integer, optional
               face to read for llc configurations (default=0)
    use_mmap : bool, optional
               Whether to read the data using a numpy.memmap

    Returns
    -------
    numpy array or memmap
    """

    # size of the data element
    nbytes = file_metadata['dtype'].itemsize
    # byte order
    file_metadata['datatype'] = file_metadata['dtype'].newbyteorder(
        file_metadata['endian'])
    # find index of variable
    idx_var = file_metadata['vars'].index(variable)

    # 1. compute offset_variable, init to zero
    offset_vars = 0
    # loop on variables before the one to read
    for jvar in np.arange(idx_var):
        # inspect its dimensions
        dims = file_metadata['dims_vars'][jvar]
        # compute the byte size of this variable
        nbytes_thisvar = 1*nbytes
        for dim in dims:
            nbytes_thisvar = nbytes_thisvar*file_metadata[dim]
        # update offset from previous variables
        offset_vars = offset_vars+nbytes_thisvar

    # 2. get dimensions of desired variable
    dims = file_metadata['dims_vars'][idx_var]
    # inquire for values of dimensions, else return 1
    nt, nz, ny, nx = [file_metadata.get(dimname, 1)
                      for dimname in ('nt', 'nz', 'ny', 'nx')]

    # 3. compute offset from previous records of current variable
    if (rec > nt-1):
        raise ValueError("time record %g greater than number of records %g" %
                         (rec, nt))
    else:
        offset_timerecords = rec * nz * ny * nx * nbytes

    # 4. compute offset from previous vertical levels of current variable
    if (lev > nz-1):
        raise ValueError("level %g is greater than number of levels %g" %
                         (lev, nz))
    else:
        offset_verticallevels = lev * ny * nx * nbytes

    # 5. compute the offset of the previous variables, records and levels
    offset = offset_vars + offset_timerecords + offset_verticallevels

    # 6. compute offset due to faces
    if file_metadata['has_faces']:
        # determin which facet the face belong to
        facet_origin = file_metadata['face_facets'][face]
        # compute the offset from previous facets
        ny_facets = np.array(file_metadata['ny_facets'])
        nyglo_facets = np.concatenate(([0], ny_facets.cumsum()[:-1]), axis=0)
        offset_facets = nyglo_facets[facet_origin] * \
            file_metadata['nx'] * nbytes
        # update offset
        offset = offset + offset_facets
        # shape if shape of the facet
        shape = (file_metadata['ny_facets'][facet_origin], nx,)
    else:
        # no need to update offset and shape is simply:
        shape = (ny, nx,)

    # check if we do a partial read of the file
    if (nt > 1) or (nz > 1) or (len(file_metadata['vars']) > 1) or \
       file_metadata['has_faces']:
        partial_read = True
    else:
        partial_read = False

    # define the order (row/column major)
    if file_metadata['has_faces']:
        # in llc, we can have either C or F
        order = file_metadata['facet_orders'][facet_origin]
    else:
        # in conventional grids, it's in C
        order = 'C'

    # 7. Do the actual read
    data_raw = read_raw_data(file_metadata['filename'],
                             file_metadata['datatype'],
                             shape, use_mmap=use_mmap, offset=offset,
                             order=order, partial_read=partial_read)

    # 8. Pad data, if needed
    data_padded_after = _pad_array(data_raw, file_metadata, face=face)

    # 9. extract the face from the facet
    if file_metadata['has_faces'] and ('face_offsets' in file_metadata):
        face_slice = slice(nx*file_metadata['face_offsets'][face],
                           nx*(file_metadata['face_offsets'][face]+1))

        data = data_padded_after[face_slice]
    else:
        data = data_padded_after

    # 10. Transpose face, if needed
    if file_metadata['has_faces'] and ('transpose_face' in file_metadata):
        if file_metadata['transpose_face'][face]:
            data = data.transpose()

    return data


def _pad_array(data, file_metadata, face=0):
    """
    Return a padded array. If input data is a numpy.memmap and no padding
    is necessary, the function preserves its type. Otherwise, the concatenate
    forces it to load into memory.

    Parameters
    ----------

    data          : numpy array or memmap
                    input data
    file_metadata : dict
                    metadata for file
    face          : int, optional
                    llc face if applicable

    Returns
    -------
    numpy.array or numpy.memmap

    """

    # Pad data before in y direction
    if 'pad_before_y' in file_metadata:
        if file_metadata['has_faces']:
            facet_origin = file_metadata['face_facets'][face]
            nypad_before = file_metadata['pad_before_y'][facet_origin]
        else:
            nypad_before = file_metadata['pad_before_y']

        pad_before = np.zeros((nypad_before, file_metadata['nx']))
        data_padded_before = np.concatenate(
            (pad_before, data), axis=0)
    else:
        data_padded_before = data

    # Pad data after in y direction
    if 'pad_after_y' in file_metadata:
        if file_metadata['has_faces']:
            facet_origin = file_metadata['face_facets'][face]
            nypad_after = file_metadata['pad_after_y'][facet_origin]
        else:
            nypad_after = file_metadata['pad_after_y']

        pad_after = np.zeros((nypad_after, file_metadata['nx']))
        data_padded_after = np.concatenate(
            (data_padded_before, pad_after), axis=0)
    else:
        data_padded_after = data_padded_before

    return data_padded_after


def get_extra_metadata(domain='llc', nx=90):
    """ 
    Return the extra_metadata dictionay for selected domains

    PARAMETERS
    ----------
    domain : str
        domain can be llc, aste, cs
    nx : int
        size of the face in the x direction

    RETURNS
    -------
    extra_metadata : dict
        all extra_metadata to handle multi-faceted grids
    """

    available_domains = ['llc', 'aste', 'cs']
    if domain not in available_domains:
        raise ValueError('not an available domain')

    # domains
    llc = {'has_faces': True, 'ny': 13*nx, 'nx': nx,
           'ny_facets': [3*nx, 3*nx, nx, 3*nx, 3*nx],
           'face_facets': [0, 0, 0, 1, 1, 1, 2, 3, 3, 3, 4, 4, 4],
           'facet_orders': ['C', 'C', 'C', 'F', 'F'],
           'face_offsets': [0, 1, 2, 0, 1, 2, 0, 0, 1, 2, 0, 1, 2],
           'transpose_face': [False, False, False,
                              False, False, False, False,
                              True, True, True, True, True, True]}

    aste = {'has_faces': True, 'ny': 5*nx, 'nx': nx,
            'ny_facets': [int(5*nx/3.), 0, nx,
                          int(2*nx/3.), int(5*nx/3.)],
            'pad_before_y': [int(1*nx/3.), 0, 0, 0, 0],
            'pad_after_y': [0, 0, 0, int(1*nx/3.), int(1*nx/3.)],
            'face_facets': [0, 0, 2, 3, 4, 4],
            'facet_orders': ['C', 'C', 'C', 'F', 'F'],
            'face_offsets': [0, 1, 0, 0, 0, 1],
            'transpose_face': [False, False, False,
                               True, True, True]}

    cs = {'has_faces': True, 'ny': nx, 'nx': nx,
          'ny_facets': [nx, nx, nx, nx, nx, nx],
          'face_facets': [0, 1, 2, 3, 4, 5],
          'facet_orders': ['F', 'F', 'F', 'F', 'F', 'F'],
          'face_offsets': [0, 0, 0, 0, 0, 0],
          'transpose_face': [False, False, False,
                             False, False, False]}

    if domain == 'llc':
        extra_metadata = llc
    elif domain == 'aste':
        extra_metadata = aste
    elif domain == 'cs':
        extra_metadata = cs

    return extra_metadata


def get_grid_from_input(gridfile, nx=None, ny=None, geometry='llc',
                        dtype=np.dtype('d'), endian='>', use_dask=False,
                        extra_metadata=None):
    """ 
    Read grid variables from grid input files, this is especially useful
    for llc and cube sphere configurations used with land tiles
    elimination. Reading the input grid files (e.g. tile00[1-5].mitgrid)
    allows to fill in the blanks of eliminated land tiles.

    PARAMETERS
    ----------
    gridfile : str
        gridfile must contain <NFACET> as wildcard (e.g. tile<NFACET>.mitgrid)
    nx : int
        size of the face in the x direction
    ny : int
        size of the face in the y direction
    geometry : str
        domain geometry can be llc, cs or carthesian not supported yet
    dtype : np.dtype
        numeric precision (single/double) of input data
    endian : string
        endianness of input data
    use_dask : bool
        use dask or not
    extra_metadata : dict
        dictionary of extra metadata, needed for llc configurations

    RETURNS
    ------- 
    grid : xarray.Dataset
        all grid variables
    """

    file_metadata = {}
    # grid variables are stored in this order
    file_metadata['fldList'] = ['XC', 'YC', 'DXF', 'DYF', 'RAC',
                                'XG', 'YG', 'DXV', 'DYU', 'RAZ',
                                'DXC', 'DYC', 'RAW', 'RAS', 'DXG', 'DYG']

    file_metadata['vars'] = file_metadata['fldList']
    dims_vars_list = []
    for var in file_metadata['fldList']:
        dims_vars_list.append(('ny', 'nx'))
    file_metadata['dims_vars'] = dims_vars_list

    # no vertical levels or time records
    file_metadata['nz'] = 1
    file_metadata['nt'] = 1

# for curvilinear non-facet grids (TO DO)
#    if nx is not None:
#        file_metadata['nx'] = nx
#    if ny is not None:
#        file_metadata['ny'] = ny
    if extra_metadata is not None:
        file_metadata.update(extra_metadata)

    # numeric representation
    file_metadata['endian'] = endian
    file_metadata['dtype'] = dtype

    if geometry == 'llc':
        nfacets = 5
        try:
            nfaces = len(file_metadata['face_facets'])
        except:
            raise ValueError('metadata must contain face_facets')
    if geometry == 'cs':  # pragma: no cover
        raise NotImplementedError("'cs' geometry is not supported yet")

    # create placeholders for data
    gridfields = {}
    for field in file_metadata['fldList']:
        gridfields.update({field: None})

    if geometry == 'llc':
        for kfacet in range(nfacets):
            # we need to adapt the metadata to the grid file
            grid_metadata = file_metadata.copy()

            fname = gridfile.replace('<NFACET>', str(kfacet+1).zfill(3))
            grid_metadata['filename'] = fname

            if file_metadata['facet_orders'][kfacet] == 'C':
                nxgrid = file_metadata['nx'] + 1
                nygrid = file_metadata['ny_facets'][kfacet] + 1
            elif file_metadata['facet_orders'][kfacet] == 'F':
                nxgrid = file_metadata['ny_facets'][kfacet] + 1
                nygrid = file_metadata['nx'] + 1

            grid_metadata.update({'nx': nxgrid, 'ny': nygrid,
                                  'has_faces': False})

            raw = read_all_variables(grid_metadata['vars'], grid_metadata,
                                     use_dask=use_dask)

            rawfields = {}
            for kfield in np.arange(len(file_metadata['fldList'])):

                rawfields.update(
                    {file_metadata['fldList'][kfield]: raw[kfield]})

            for field in file_metadata['fldList']:
                # symetrize
                tmp = rawfields[field][:, :, :-1, :-1].squeeze()
                # transpose
                if grid_metadata['facet_orders'][kfacet] == 'F':
                    tmp = tmp.transpose()

                for face in np.arange(nfaces):
                    # identify faces that need to be filled
                    if grid_metadata['face_facets'][face] == kfacet:
                        # get offset of face from facet
                        offset = file_metadata['face_offsets'][face]
                        nx = file_metadata['nx']
                        # pad data, if needed (would trigger eager data eval)
                        # needs a new array not to pad multiple times
                        padded = _pad_array(tmp, file_metadata, face=face)
                        # extract the data
                        dataface = padded[offset*nx:(offset+1)*nx, :]
                        # transpose, if needed
                        if file_metadata['transpose_face'][face]:
                            dataface = dataface.transpose()
                        # assign values
                        dataface = dsa.stack([dataface], axis=0)
                        if face == 0:
                            gridfields[field] = dataface
                        else:
                            gridfields[field] = dsa.concatenate(
                                [gridfields[field], dataface], axis=0)

    elif geometry == 'cs':  # pragma: no cover
        raise NotImplementedError("'cs' geometry is not supported yet")
        pass

    # create the dataset
    if geometry in ['llc', 'cs']:
        grid = xr.Dataset({'XC':  (['face', 'j', 'i'],     gridfields['XC']),
                           'YC':  (['face', 'j', 'i'],     gridfields['YC']),
                           'DXF': (['face', 'j', 'i'],     gridfields['DXF']),
                           'DYF': (['face', 'j', 'i'],     gridfields['DYF']),
                           'RAC': (['face', 'j', 'i'],     gridfields['RAC']),
                           'XG':  (['face', 'j_g', 'i_g'], gridfields['XG']),
                           'YG':  (['face', 'j_g', 'i_g'], gridfields['YG']),
                           'DXV': (['face', 'j', 'i'],     gridfields['DXV']),
                           'DYU': (['face', 'j', 'i'],     gridfields['DYU']),
                           'RAZ': (['face', 'j_g', 'i_g'], gridfields['RAZ']),
                           'DXC': (['face', 'j', 'i_g'],   gridfields['DXC']),
                           'DYC': (['face', 'j_g', 'i'],   gridfields['DYC']),
                           'RAW': (['face', 'j', 'i_g'],   gridfields['RAW']),
                           'RAS': (['face', 'j_g', 'i'],   gridfields['RAS']),
                           'DXG': (['face', 'j_g', 'i'],   gridfields['DXG']),
                           'DYG': (['face', 'j', 'i_g'],   gridfields['DYG'])
                           },
                          coords={'i': (['i'], np.arange(file_metadata['nx'])),
                                  'j': (['j'], np.arange(file_metadata['nx'])),
                                  'i_g': (['i_g'],
                                          np.arange(file_metadata['nx'])),
                                  'j_g': (['j_g'],
                                          np.arange(file_metadata['nx'])),
                                  'face': (['face'], np.arange(nfaces))
                                  }
                          )
    else:  # pragma: no cover
        grid = xr.Dataset({'XC':  (['j', 'i'],     gridfields['XC']),
                           'YC':  (['j', 'i'],     gridfields['YC']),
                           'DXF': (['j', 'i'],     gridfields['DXF']),
                           'DYF': (['j', 'i'],     gridfields['DYF']),
                           'RAC': (['j', 'i'],     gridfields['RAC']),
                           'XG':  (['j_g', 'i_g'], gridfields['XG']),
                           'YG':  (['j_g', 'i_g'], gridfields['YG']),
                           'DXV': (['j', 'i'],     gridfields['DXV']),
                           'DYU': (['j', 'i'],     gridfields['DYU']),
                           'RAZ': (['j_g', 'i_g'], gridfields['RAZ']),
                           'DXC': (['j', 'i_g'],   gridfields['DXC']),
                           'DYC': (['j_g', 'i'],   gridfields['DYC']),
                           'RAW': (['j', 'i_g'],   gridfields['RAW']),
                           'RAS': (['j_g', 'i'],   gridfields['RAS']),
                           'DXG': (['j_g', 'i'],   gridfields['DXG']),
                           'DYG': (['j', 'i_g'],   gridfields['DYG'])
                           },
                          coords={'i': (['i'], np.arange(file_metadata['nx'])),
                                  'j': (['j'], np.arange(file_metadata['ny'])),
                                  'i_g': (['i_g'],
                                          np.arange(file_metadata['nx'])),
                                  'j_g': (['j_g'],
                                          np.arange(file_metadata['ny']))
                                  }
                          )

    return grid


########## WRITING BINARIES #############################

def find_concat_dim_facet(da, facet, extra_metadata):
    """ In llc grids, find along which horizontal dimension to concatenate
    facet between i, i_g and j, j_g. If the order of the facet is F, concat
    along i or i_g. If order is C, concat along j or j_g. Also return
    horizontal dim not to concatenate

    PARAMETERS
    ----------
    da : xarray.DataArray
        xmitgcm llc data array
    facet : int
        facet number
    extra_metadata : dict
        dict of extra_metadata from get_extra_metadata

    RETURNS
    -------
    concat_dim, nonconcat_dim : str, str
        names of the dimensions for concatenation or not

    """
    order = extra_metadata['facet_orders'][facet]
    if order == 'C':
        possible_concat_dims = ['j', 'j_g']
    elif order == 'F':
        possible_concat_dims = ['i', 'i_g']

    concat_dim = find_concat_dim(da, possible_concat_dims)

    # we also need to other horizontal dimension for vector indexing
    all_dims = list(da.dims)
    # discard face
    all_dims.remove('face')
    # remove the concat_dim to find horizontal non_concat dimension
    all_dims.remove(concat_dim)
    non_concat_dim = all_dims[0]
    return concat_dim, non_concat_dim


def find_concat_dim(da, possible_concat_dims):
    """ look for available dimensions in dataaray and pick the one
    from a list of candidates

    PARAMETERS
    ----------
    da : xarray.DataArray
        xmitgcm llc data array
    possible_concat_dims : list
        list of potential dims

    RETURNS
    -------
    out : str
        dimension on which to concatenate

    """
    out = None
    for d in possible_concat_dims:
        if d in da.dims:
            out = d
    return out


def rebuild_llc_facets(da, extra_metadata):
    """ For LLC grids, rebuilds facets from a xmitgcm dataarray and
    store into a dictionary

    PARAMETERS
    ----------
    da : xarray.DataArray
        xmitgcm llc data array
    extra_metadata : dict
        dict of extra_metadata from get_extra_metadata

    RETURNS
    -------
    facets : dict
        all facets data in xarray.DataArray form packed into a dictionary

    """

    nfacets = len(extra_metadata['facet_orders'])
    nfaces = len(extra_metadata['face_facets'])
    facets = {}

    # rebuild the facets (with padding if present)
    for kfacet in range(nfacets):
        facets.update({'facet' + str(kfacet): None})

        concat_dim, non_concat_dim = find_concat_dim_facet(
            da, kfacet, extra_metadata)

        for kface in range(nfaces):
            # concatenate faces back into facets
            if extra_metadata['face_facets'][kface] == kfacet:
                if extra_metadata['face_offsets'][kface] == 0:
                    # first face of facet
                    tmp = da.sel(face=kface)
                else:
                    # any other face needs to be concatenated
                    newface = da.sel(face=kface)
                    tmp = xr.concat([facets['facet' + str(kfacet)],
                                     newface], dim=concat_dim)

                facets['facet' + str(kfacet)] = tmp

    # if present, remove padding from facets
    for kfacet in range(nfacets):

        concat_dim, non_concat_dim = find_concat_dim_facet(
            da, kfacet, extra_metadata)

        # remove pad before
        if 'pad_before_y' in extra_metadata:
            pad = extra_metadata['pad_before_y'][kfacet]
            # padded array
            padded = facets['facet' + str(kfacet)]

            if pad != 0:
                # we need to relabel the grid cells
                ng = len(padded[concat_dim].values)
                padded[concat_dim] = np.arange(ng)
                # select index from non-padded array
                unpadded_bef = padded.isel({concat_dim: range(pad, ng)})
            else:
                unpadded_bef = padded

            facets['facet' + str(kfacet)] = unpadded_bef

        # remove pad after
        if 'pad_after_y' in extra_metadata:
            pad = extra_metadata['pad_after_y'][kfacet]
            # padded array
            padded = facets['facet' + str(kfacet)]

            if pad != 0:
                # we need to relabel the grid cells
                ng = len(padded[concat_dim].values)
                padded[concat_dim] = np.arange(ng)
                # select index from non-padded array
                last = ng-pad
                unpadded_aft = padded.isel({concat_dim: range(last)})
            else:
                unpadded_aft = padded

            facets['facet' + str(kfacet)] = unpadded_aft

    return facets


def llc_facets_3d_spatial_to_compact(facets, dimname, extra_metadata):
    """ Write in compact form a list of 3d facets

    PARAMETERS
    ----------
    facets : dict
        dict of xarray.dataarrays for the facets
    extra_metadata : dict
        extra_metadata from get_extra_metadata

    RETURNS
    -------
    flatdata : numpy.array
        all the data in vector form
    """

    nz = len(facets['facet0'][dimname])
    nfacets = len(facets)
    flatdata = np.array([])

    for kz in range(nz):
        # rebuild the dict
        tmpdict = {}
        for kfacet in range(nfacets):
            this_facet = facets['facet' + str(kfacet)]
            if this_facet is not None:
                tmpdict['facet' + str(kfacet)] = this_facet.isel(k=kz)
            else:
                tmpdict['facet' + str(kfacet)] = None
        # concatenate all 2d arrays
        compact2d = llc_facets_2d_to_compact(tmpdict, extra_metadata)
        flatdata = np.concatenate([flatdata, compact2d])

    return flatdata


def llc_facets_2d_to_compact(facets, extra_metadata):
    """ Write in compact form a list of 2d facets

    PARAMETERS
    ----------
    facets: dict
        dict of xarray.dataarrays for the facets
    extra_metadata: dict
        extra_metadata from get_extra_metadata

    RETURNS
    -------
    flatdata : numpy.array
        all the data in vector form
    """

    flatdata = np.array([])
    # loop over facets
    for kfacet in range(len(facets)):
        if facets['facet' + str(kfacet)] is not None:
            tmp = np.reshape(facets['facet' + str(kfacet)].values, (-1))
            flatdata = np.concatenate([flatdata, tmp])

    return flatdata


def write_to_binary(flatdata, fileout, dtype=np.dtype('f')):
    """ write data in binary file

    PARAMETERS
    ----------
    flatdata: numpy.array
        vector of data to write
    fileout: str
        output file name
    dtype: np.dtype
        single/double precision

    RETURNS
    -------
    None
    """
    # write data to binary files
    fid = open(fileout, "wb")
    tmp = flatdata.astype(dtype)
    if sys.byteorder == 'little':
        tmp = tmp.byteswap(True)
    fid.write(tmp.tobytes())
    fid.close()
    return None



import os

from .llcmodel import BaseLLCModel
from . import stores


def _requires_pleiades(func):
    def wrapper(*args, **kwargs):
        # is there a better choice
        test_path = '/home6/dmenemen'
        if not os.path.exists(test_path):
            raise OSError("Can't find %s. We must not be on Pleiades." % test_path)
        func(*args, **kwargs)
    return wrapper

def _requires_sverdrup(func):
    def wrapper(*args,**kwargs):
        test_path = '/scratch2/heimbach'
        if not os.path.exists(test_path):
            raise OSError("Can't find %s. We must not be on Sverdrup." % test_path)
        func(*args, **kwargs)
    return wrapper


def _make_http_filesystem():
    import fsspec
    from fsspec.implementations.http import HTTPFileSystem
    return HTTPFileSystem()

class LLC90Model(BaseLLCModel):
    nx = 90
    nz = 50
    delta_t = 3600
    time_units = 'seconds since 1948-01-01 12:00:00'
    calendar = 'gregorian'


class LLC2160Model(BaseLLCModel):
    nx = 2160
    nz = 90
    delta_t = 45
    iter_start = 92160
    iter_stop = 1586400 + 1
    iter_step = 80
    time_units='seconds since 2011-01-17'
    calendar = 'gregorian'
    varnames = ['Eta', 'KPPhbl', 'oceFWflx', 'oceQnet', 'oceQsw', 'oceSflux',
                'oceTAUX', 'oceTAUY', 'PhiBot', 'Salt', 'SIarea', 'SIheff',
                'SIhsalt', 'SIhsnow', 'SIuice', 'SIvice', 'Theta', 'U', 'V', 'W']
    grid_varnames = ['AngleCS','AngleSN','Depth',
                     'DRC','DRF',
                     'DXC','DXF','DXG',
                     'DYC','DYF','DYG',
                     'hFacC','hFacS','hFacW','PHrefC','PHrefF',
                     'RAC','RAS','RAW',
                     'RC','RF','RhoRef','rLowC','rLowS','rLowW',
                     'rSurfC','rSurfS','rSurfW','XC','YC',
                     'RAZ','XG','YG','DXV','DYU']
    mask_override = {'oceTAUX': 'c', 'oceTAUY': 'c'}


class LLC4320Model(BaseLLCModel):
    nx = 4320
    nz = 90
    delta_t = 25
    iter_start = 10368
    iter_stop = 1495152 + 1
    iter_step = 144


    time_units='seconds since 2011-09-10'
    calendar = 'gregorian'
    varnames = ['Eta', 'KPPhbl', 'oceFWflx', 'oceQnet', 'oceQsw', 'oceSflux',
                'oceTAUX', 'oceTAUY', 'PhiBot', 'Salt', 'SIarea', 'SIheff',
                'SIhsalt', 'SIhsnow', 'SIuice', 'SIvice', 'Theta', 'U', 'V', 'W']
    grid_varnames = ['AngleCS','AngleSN','Depth',
                     'DRC','DRF',
                     'DXC','DXF','DXG',
                     'DYC','DYF','DYG',
                     'hFacC','hFacS','hFacW','PHrefC','PHrefF',
                     'RAC','RAS','RAW','RC','RF',
                     'RhoRef','XC','YC','RAZ','XG','YG','DXV','DYU']
    mask_override = {'oceTAUX': 'c', 'oceTAUY': 'c'}

class ASTE270Model(BaseLLCModel):
    nface = 6
    nx = 270
    nz = 50
    domain = 'aste'
    pad_before = [90, 0, 0, 0, 0]
    pad_after = [0, 0, 0, 90, 90]
    delta_t = 600
    time_units='seconds since 2002-01-01'
    calendar = 'gregorian'

    iters=[4464, 8496, 12960, 17280, 21744,
           26064, 30528, 34992, 39312, 43776,
           48096, 52560, 57024, 61056, 65520,
           69840, 74304, 78624, 83088, 87552,
           91872, 96336, 100656, 105120, 109584,
           113760, 118224, 122544, 127008, 131328,
           135792, 140256, 144576, 149040, 153360,
           157824, 162288, 166320, 170784, 175104,
           179568, 183888, 188352, 192816, 197136,
           201600, 205920, 210384, 214848, 218880,
           223344, 227664, 232128, 236448, 240912,
           245376, 249696, 254160, 258480, 262944,
           267408, 271440, 275904, 280224, 284688,
           289008, 293472, 297936, 302256, 306720,
           311040, 315504, 319968, 324144, 328608,
           332928, 337392, 341712, 346176, 350640,
           354960, 359424, 363744, 368208, 372672,
           376704, 381168, 385488, 389952, 394272,
           398736, 403200, 407520, 411984, 416304,
           420768, 425232, 429264, 433728, 438048,
           442512, 446832, 451296, 455760, 460080,
           464544, 468864, 473328, 477792, 481824,
           486288, 490608, 495072, 499392, 503856,
           508320, 512640, 517104, 521424, 525888,
           530352, 534528, 538992, 543312, 547776,
           552096, 556560, 561024, 565344, 569808,
           574128, 578592, 583056, 587088, 591552,
           595872, 600336, 604656, 609120, 613584,
           617904, 622368, 626688, 631152, 635616,
           639648, 644112, 648432, 652896, 657216,
           661680, 666144, 670464, 674928, 679248,
           683712, 688176, 692208, 696672, 700992,
           705456, 709776, 714240, 718704, 723024,
           727488, 731808, 736272, 740736, 744912,
           749376, 753696, 758160, 762480, 766944,
           771408, 775728, 780192, 784512, 788976,
           793440, 797472, 801936, 806256, 810720,
           815040, 819504, 823968, 828288, 832752,
           837072, 841536, 841544]
    varnames = ['ADVr_SLT', 'ADVr_TH',  'ADVxHEFF', 'ADVxSNOW', 'ADVx_SLT',
                'ADVx_TH',  'ADVyHEFF', 'ADVySNOW', 'ADVy_SLT', 'ADVy_TH',
                'DETADT2',  'DFrE_SLT', 'DFrE_TH',  'DFrI_SLT', 'DFrI_TH',
                'DFxEHEFF', 'DFxESNOW', 'DFxE_SLT', 'DFxE_TH',  'DFyEHEFF',
                'DFyESNOW', 'DFyE_SLT', 'DFyE_TH',  'ETAN',     'ETANSQ',
                'GM_PsiX',  'GM_PsiY',  'KPPg_SLT', 'KPPg_TH',  'MXLDEPTH',
                'PHIBOT',   'SALT',     'SFLUX',    'SIaaflux', 'SIacSubl',
                'SIarea',   'SIatmFW',  'SIatmQnt', 'SIheff',   'SIhsnow',
                'SIsnPrcp', 'SItflux',  'SIuice',   'SIvice',   'SRELAX',
                'TFLUX',    'THETA',    'TRELAX',   'UVELMASS', 'VVELMASS',
                'WSLTMASS', 'WTHMASS',  'WVELMASS', 'oceFWflx', 'oceQnet',
                'oceQsw',   'oceSPDep', 'oceSPflx', 'oceSPtnd', 'oceSflux',
                'oceTAUX',  'oceTAUY',  'sIceLoad']

    grid_varnames = ['AngleCS', 'AngleSN',   'DRC',       'DRF',       'DXC',
                     'DXG',     'DYC',       'DYG',       'Depth',     'PHrefC',
                     'PHrefF',  'RAC',       'RAS',       'RAW',       'RAZ',
                     'RC',      'RF',        'RhoRef',    'XC',        'XG',
                     'YC',      'YG',        'hFacC',     'hFacS',     'hFacW',
                     'maskC',   'maskCtrlC', 'maskCtrlS', 'maskCtrlW', 'maskInC',
                     'maskInS', 'maskInW',   'maskS',     'maskW']

    dtype={"ADVr_SLT":">f8", "ADVr_TH":">f8", "ADVxHEFF":">f8",
           "ADVxSNOW":">f8", "ADVx_SLT":">f8", "ADVx_TH":">f8",
           "ADVyHEFF":">f8", "ADVySNOW":">f8", "ADVy_SLT":">f8",
           "ADVy_TH":">f8", "AngleCS":">f8", "AngleSN":">f8",
           "DETADT2":">f4", "DFrE_SLT":">f8", "DFrE_TH":">f8",
           "DFrI_SLT":">f8", "DFrI_TH":">f8", "DFxEHEFF":">f8",
           "DFxESNOW":">f8", "DFxE_SLT":">f8", "DFxE_TH":">f8",
           "DFyEHEFF":">f8", "DFyESNOW":">f8", "DFyE_SLT":">f8",
           "DFyE_TH":">f8", "DRC":">f8", "DRF":">f8",
           "DXC":">f8", "DXG":">f8", "DYC":">f8",
           "DYG":">f8", "Depth":">f8", "ETAN":">f4",
           "ETANSQ":">f4", "GM_PsiX":">f4", "GM_PsiY":">f4",
           "KPPg_SLT":">f8", "KPPg_TH":">f8", "MXLDEPTH":">f4",
           "PHIBOT":">f4", "PHrefC":">f8", "PHrefF":">f8",
           "RAC":">f8", "RAS":">f8", "RAW":">f8",
           "RAZ":">f8", "RC":">f8", "RF":">f8",
           "RhoRef":">f8", "SALT":">f4", "SFLUX":">f8",
           "SIaaflux":">f8", "SIacSubl":">f8", "SIarea":">f4",
           "SIatmFW":">f8", "SIatmQnt":">f8", "SIheff":">f4",
           "SIhsnow":">f4", "SIsnPrcp":">f8", "SItflux":">f8",
           "SIuice":">f4", "SIvice":">f4", "SRELAX":">f8",
           "TFLUX":">f8", "THETA":">f4", "TRELAX":">f8",
           "UVELMASS":">f8", "VVELMASS":">f8", "WSLTMASS":">f8",
           "WTHMASS":">f8", "WVELMASS":">f8", "XC":">f8",
           "XG":">f8", "YC":">f8", "YG":">f8",
           "hFacC":">f8", "hFacS":">f8", "hFacW":">f8",
           "maskC":">f8", "maskCtrlC":">f8", "maskCtrlS":">f8",
           "maskCtrlW":">f8", "maskInC":">f8", "maskInS":">f8",
           "maskInW":">f8", "maskS":">f8", "maskW":">f8",
           "oceFWflx":">f8", "oceQnet":">f8", "oceQsw":">f8",
           "oceSPDep":">f4", "oceSPflx":">f8", "oceSPtnd":">f8",
           "oceSflux":">f8", "oceTAUX":">f4", "oceTAUY":">f4",
           "sIceLoad":">f4"}

class ECCOPortalLLC2160Model(LLC2160Model):

    def __init__(self):
        fs = _make_http_filesystem()
        base_path = 'https://data.nas.nasa.gov/ecco/download_data.php?file=/eccodata/llc_2160/compressed'
        grid_path = 'https://data.nas.nasa.gov/ecco/download_data.php?file=/eccodata/llc_2160/grid'
        mask_path = 'https://storage.googleapis.com/pangeo-ecco/llc/masks/llc_2160_masks.zarr/'
        store = stores.NestedStore(fs, base_path=base_path, mask_path=mask_path,
                                   grid_path=grid_path, shrunk=True, join_char='/')
        super(ECCOPortalLLC2160Model, self).__init__(store)


class ECCOPortalLLC4320Model(LLC4320Model):

    def __init__(self):
        fs = _make_http_filesystem()
        base_path = 'https://data.nas.nasa.gov/ecco/download_data.php?file=/eccodata/llc_4320/compressed'
        grid_path = 'https://data.nas.nasa.gov/ecco/download_data.php?file=/eccodata/llc_4320/grid'
        mask_path = 'https://storage.googleapis.com/pangeo-ecco/llc/masks/llc_4320_masks.zarr/'
        store = stores.NestedStore(fs, base_path=base_path, mask_path=mask_path,
                                   grid_path=grid_path, shrunk=True, join_char='/')
        super(ECCOPortalLLC4320Model, self).__init__(store)


class PleiadesLLC2160Model(LLC2160Model):

    @_requires_pleiades
    def __init__(self):
        from fsspec.implementations.local import LocalFileSystem
        fs = LocalFileSystem()
        base_path = '/home6/dmenemen/llc_2160/compressed'
        mask_path = '/nobackup/rpaberna/llc/masks/llc_2160_masks.zarr'
        store = stores.NestedStore(fs, base_path=base_path, mask_path=mask_path,
                                   shrunk=True)
        super(PleiadesLLC2160Model, self).__init__(store)


class PleiadesLLC4320Model(LLC4320Model):

    @_requires_pleiades
    def __init__(self):
        from fsspec.implementations.local import LocalFileSystem
        fs = LocalFileSystem()
        base_path = '/home6/dmenemen/llc_4320/compressed'
        mask_path = '/nobackup/rpaberna/llc/masks/llc_4320_masks.zarr'
        store = stores.NestedStore(fs, base_path=base_path, mask_path=mask_path,
                                   shrunk=True)
        super(PleiadesLLC4320Model, self).__init__(store)

class CRIOSPortalASTE270Model(ASTE270Model):

    def __init__(self):
        fs = _make_http_filesystem()
        base_path = 'https://aste-release1.s3.us-east-2.amazonaws.com/diags'
        grid_path = 'https://aste-release1.s3.us-east-2.amazonaws.com/grid'
        mask_path = 'https://aste-release1.s3.us-east-2.amazonaws.com/masks.zarr'
        store = stores.NestedStore(fs, base_path=base_path, grid_path=grid_path,
                                   mask_path=mask_path,
                                   shrunk=True, join_char='/')

        super(CRIOSPortalASTE270Model, self).__init__(store)

class SverdrupASTE270Model(ASTE270Model):

    @_requires_sverdrup
    def __init__(self):
        from fsspec.implementations.local import LocalFileSystem
        fs = LocalFileSystem()
        base_path = '/scratch2/shared/aste-release1/diags'
        grid_path = '/scratch2/shared/aste-release1/grid'
        mask_path = '/scratch2/shared/aste-release1/masks.zarr'
        store = stores.NestedStore(fs, base_path=base_path, grid_path=grid_path,
                                   mask_path=mask_path,
                                   shrunk=True, join_char='/')

        super(SverdrupASTE270Model, self).__init__(store)



import numpy as np
import dask
import dask.array as dsa
from dask.base import tokenize, normalize_token
import xarray as xr
import warnings

from .duck_array_ops import concatenate
from .shrunk_index import all_index_data
from ..variables import dimensions

def _get_grid_metadata():
    # keep this separate from get_var_metadata
    # because grid stuff is weird
    from ..mds_store import _get_all_grid_variables
    from ..variables import extra_grid_variables, vertical_coordinates

    # get grid info
    grid_vars = _get_all_grid_variables('llc')
    grid_vars.update(extra_grid_variables)

    # make dictionary with keys as filenames
    grid_metadata = {}
    for key,val in grid_vars.items():
        # masks use hFac filename to be computed in mds_store
        if 'filename' in val and key[:4]!='mask':
            val.update({'real_name':key})
            grid_metadata[val['filename']] = val
        else:
            grid_metadata[key] = val

    # force RF to point to Zp1, deal with this manually..
    grid_metadata['RF']=vertical_coordinates['Zp1']
    grid_metadata['RF']['real_name'] = 'Zp1'
    for zv in ['Zu','Zl']:
        grid_metadata[zv] = vertical_coordinates[zv]

    return grid_metadata

def _get_var_metadata():
    # The LLC run data comes with zero metadata. So we import metadata from
    # the xmitgcm package.
    from ..variables import state_variables, package_state_variables
    from ..utils import parse_available_diagnostics
    from ..default_diagnostics import diagnostics
    from io import StringIO

    diag_file = StringIO(diagnostics)
    available_diags = parse_available_diagnostics(diag_file)
    var_metadata = state_variables.copy()
    var_metadata.update(package_state_variables)
    var_metadata.update(available_diags)

    # even the file names from the LLC data differ from standard MITgcm output
    aliases = {'Eta': 'ETAN', 'PhiBot': 'PHIBOT', 'Salt': 'SALT',
               'Theta': 'THETA'}
    for a, b in aliases.items():
        var_metadata[a] = var_metadata[b]

    # add grid metadata
    var_metadata.update(_get_grid_metadata())

    return var_metadata

_VAR_METADATA = _get_var_metadata()

def _is_vgrid(vname):
    # check for 1d, vertical grid variables
    dims = _VAR_METADATA[vname]['dims']
    return len(dims)==1 and dims[0][0]=='k'

def _get_variable_point(vname, mask_override):
    # fix for https://github.com/MITgcm/xmitgcm/issues/191
    if vname in mask_override:
        return mask_override[vname]
    dims = _VAR_METADATA[vname]['dims']
    if 'i' in dims and 'j' in dims:
        point = 'c'
    elif 'i_g' in dims and 'j' in dims:
        point = 'w'
    elif 'i' in dims and 'j_g' in dims:
        point = 's'
    elif 'i_g' in dims and 'j_g' in dims:
        raise ValueError("Don't have masks for corner points!")
    else:
        raise ValueError("Variable `%s` is not a horizontal variable." % vname)
    return point

def _get_scalars_and_vectors(varnames, type):

    for vname in varnames:
        if vname not in _VAR_METADATA:
            raise ValueError("Varname `%s` not found in metadata." % vname)

    if type != 'latlon':
        return varnames, []

    scalars = []
    vector_pairs = []
    for vname in varnames:
        meta = _VAR_METADATA[vname]
        try:
            mate = meta['attrs']['mate']
            if mate not in varnames:
                raise ValueError("Vector pairs are required to create "
                                 "latlon type datasets. Varname `%s` is "
                                 "missing its vector mate `%s`"
                                 % vname, mate)
            vector_pairs.append((vname, mate))
            varnames.remove(mate)
        except KeyError:
            scalars.append(vname)

def _decompress(data, mask, dtype):
    data_blank = np.full_like(mask, np.nan, dtype=dtype)
    data_blank[mask] = data
    data_blank.shape = mask.shape
    return data_blank

def _pad_facet(data,facet_shape,reshape,pad_before,pad_after,dtype):
    """add padding to facet data that are irregularly shaped, returning
    data of size ``facet_shape`` to make equally sized faces.
    See
    https://xmitgcm.readthedocs.io/en/latest/llcreader.html#aste-release-1-example
    for an example

    Parameters
    ----------
    data : array like
        data to be reshaped
    facet_shape : tuple
        "expected" facet shape determined by the _facet_shape function
    reshape : bool
        whether to reshape each face, True if facet is "rotated".
        This determines the axis along which pad_before or pad_after refers to
    pad_before, pad_after : int
        size of padding added to either the i or j dimension where
        before vs after determines ordering of: data-then-pad or pad-then-data
    dtype : numpy.dtype
        Datatype of the data

    Returns
    -------
    padded_data : array like
        which has shape = facet_shape, with nan's for padding
    """
    pre_shape=list(facet_shape)
    pad_shape=list(facet_shape)
    if reshape:
        concat_axis = -1
        pad_shape[concat_axis] = pad_after
        pre_shape[concat_axis] -= pad_after
        padded= [data,np.full(pad_shape,np.nan,dtype=dtype)]
    else:
        concat_axis = -2
        pad_shape[concat_axis] = pad_before
        pre_shape[concat_axis] -= pad_before
        padded = [np.full(pad_shape,np.nan,dtype=dtype),data]

    data.shape = pre_shape
    return concatenate(padded,axis=concat_axis)

def _facet_strides(nfaces):
    if nfaces == 13:
        return ((0,3), (3,6), (6,7), (7,10), (10,13))
    elif nfaces == 6:
        return ((0,2), (2,2), (2,3), (3,4), (4,6))
    else:
        raise TypeError(f'Unexpected nfaces {nfaces} for _facet_strides')

# whether to reshape each face
_facet_reshape = (False, False, False, True, True)
_nfacets = 5

def _uncompressed_facet_index(nfacet, nside, nfaces):
    face_size = nside**2
    start = _facet_strides(nfaces)[nfacet][0] * face_size
    end = _facet_strides(nfaces)[nfacet][1] * face_size
    return start, end

def _facet_shape(nfacet, nside, nfaces):
    facet_length = _facet_strides(nfaces)[nfacet][1] - _facet_strides(nfaces)[nfacet][0]
    if _facet_reshape[nfacet]:
        facet_shape = (1, nside, facet_length*nside)
    else:
        facet_shape = (1, facet_length*nside, nside)
    return facet_shape

def _facet_to_faces(data, nfacet, nfaces):
    shape = data.shape
    # facet dimension
    nf, ny, nx = shape[-3:]
    other_dims = shape[:-3]
    assert nf == 1
    facet_length = _facet_strides(nfaces)[nfacet][1] - _facet_strides(nfaces)[nfacet][0]
    if _facet_reshape[nfacet]:
        new_shape = other_dims +  (ny, facet_length, nx / facet_length) if facet_length > 0 else 0
        data_rs = data.reshape(new_shape)
        data_rs = np.moveaxis(data_rs, -2, -3) # dask-safe
    else:
        new_shape = other_dims + (facet_length, ny / facet_length, nx) if facet_length > 0 else 0
        data_rs = data.reshape(new_shape) if facet_length>0 else None
    return data_rs

def _facets_to_faces(facets, nfaces):
    all_faces = []
    for nfacet, data_facet in enumerate(facets):
        data_rs = _facet_to_faces(data_facet, nfacet, nfaces)
        if data_rs is not None:
            all_faces.append(data_rs)
    return concatenate(all_faces, axis=-3)

def _faces_to_facets(data, nfaces, facedim=-3):
    assert data.shape[facedim] == nfaces
    facets = []
    for nfacet, (strides, reshape) in enumerate(zip(_facet_strides(nfaces), _facet_reshape)):
        face_data = [data[(...,) + (slice(nface, nface+1), slice(None), slice(None))]
                     for nface in range(*strides)]
        if reshape:
            concat_axis = facedim + 2
        else:
            concat_axis = facedim + 1
        # todo: use duck typing for concat
        facet_data = concatenate(face_data, axis=concat_axis) if len(face_data)!=0 else np.array([])
        facets.append(facet_data)
    return facets


def _rotate_scalar_facet(facet):
    facet_transposed = np.moveaxis(facet, -1, -2)
    facet_rotated = np.flip(facet_transposed, -2)
    return facet_rotated


def _facets_to_latlon_scalar(all_facets):
    rotated = (all_facets[:2]
               + [_rotate_scalar_facet(facet) for facet in all_facets[-2:]])
    # drop facet dimension
    rotated = [r[..., 0, :, :] for r in rotated]
    return concatenate(rotated, axis=-1)


def _faces_to_latlon_scalar(data, nfaces):
    data_facets = _faces_to_facets(data, nfaces)
    return _facets_to_latlon_scalar(data_facets)


# dask's pad function doesn't work
# it does weird things to non-pad dimensions
# need to roll our own
def shift_and_pad(a,left=True):
    if left:
        a_shifted = a[..., 1:]
        pad_array = dsa.zeros_like(a[..., -2:-1])
        return concatenate([a_shifted, pad_array], axis=-1)
    else:
        a_shifted = a[..., :-1]
        pad_array = dsa.zeros_like(a[..., 0:1])
        return concatenate([pad_array, a_shifted], axis=-1)

def transform_v_to_u(facet):
    return _rotate_scalar_facet(facet)

def transform_u_to_v(facet, metric=False):
    # "shift" u component by 1 pixel
    pad_width = (facet.ndim - 1) * (None,) + ((1, 0),)
    #facet_padded = dsa.pad(facet[..., 1:], pad_width, 'constant')
    facet_padded = shift_and_pad(facet)
    assert facet.shape == facet_padded.shape
    facet_rotated = _rotate_scalar_facet(facet_padded)
    if not metric:
        facet_rotated = -facet_rotated
    return facet_rotated

def _facets_to_latlon_vector(facets_u, facets_v, metric=False):
    # need to pad the rotated v values
    ndim = facets_u[0].ndim
    # second-to-last axis is the one to pad, plus a facet axis
    assert ndim >= 3

    # drop facet dimension
    facets_u_drop = [f[..., 0, :, :] for f in facets_u]
    facets_v_drop = [f[..., 0, :, :] for f in facets_v]

    u_rot = (facets_u_drop[:2]
             + [transform_v_to_u(facet) for facet in facets_v_drop[-2:]])
    v_rot = (facets_v_drop[:2]
             + [transform_u_to_v(facet, metric) for facet in facets_u_drop[-2:]])

    u = concatenate(u_rot, axis=-1)
    v = concatenate(v_rot, axis=-1)
    return u, v

def _faces_to_latlon_vector(u_faces, v_faces, nfaces, metric=False):
    u_facets = _faces_to_facets(u_faces, nfaces)
    v_facets = _faces_to_facets(v_faces, nfaces)
    u, v = _facets_to_latlon_vector(u_facets, v_facets, metric=metric)
    return u, v

def _drop_facedim(dims):
    dims = list(dims)
    dims.remove('face')
    return dims

def _add_face_to_dims(dims):
    new_dims = dims.copy()
    if 'j' in dims:
        j_dim = dims.index('j')
        new_dims.insert(j_dim, 'face')
    elif 'j_g' in dims:
        j_dim = dims.index('j_g')
        new_dims.insert(j_dim, 'face')
    return new_dims

def _faces_coords_to_latlon(ds):
    coords = ds.reset_coords().coords.to_dataset()
    ifac = 4
    jfac = 3
    dim_coords = {}
    for vname in coords.coords:
        if vname[0] == 'i':
            data = np.arange(ifac * coords.dims[vname])
        elif vname[0] == 'j':
            data = np.arange(jfac * coords.dims[vname])
        else:
            data = coords[vname].data
        var = xr.Variable(ds[vname].dims, data, ds[vname].attrs)
        dim_coords[vname] = var
    return xr.Dataset(dim_coords)

def faces_dataset_to_latlon(ds, metric_vector_pairs=[('dxC', 'dyC'), ('dyG', 'dxG')]):
    """Transform a 13-face LLC xarray Dataset into a rectancular grid,
    discarding the Arctic.

    Parameters
    ----------
    ds : xarray.Dataset
        A 13-face LLC dataset
    metric_vector_pairs : list, optional
        Pairs of variables that are positive-definite metrics located at grid
        edges.

    Returns
    -------
    out : xarray.Dataset
        Transformed rectangular dataset
    """

    coord_vars = list(ds.coords)
    ds_new = _faces_coords_to_latlon(ds)

    vector_pairs = []
    scalars = []
    vnames = list(ds.reset_coords().variables)

    for vname in vnames:
        try:
            mate = ds[vname].attrs['mate']
        except KeyError:
            mate = None

        # Raises an exception if the mate of a variable in vnames is missing.
        if mate is not None:
            vector_pairs.append((vname, mate))
            try:
                vnames.remove(mate)
            except ValueError:
                msg = 'If {} in varnames, {} must also be in varnames'.format(vname, mate)
                raise ValueError(msg)

    all_vector_components = [inner for outer in (vector_pairs + metric_vector_pairs)
                             for inner in outer]
    scalars = [vname for vname in vnames if vname not in all_vector_components]
    data_vars = {}

    for vname in scalars:
        if vname=='face' or vname in ds_new:
            continue
        if 'face' in ds[vname].dims:
            data = _faces_to_latlon_scalar(ds[vname].data,nfaces=len(ds['face']))
            dims = _drop_facedim(ds[vname].dims)
        else:
            data = ds[vname].data
            dims = ds[vname].dims
        data_vars[vname] = xr.Variable(dims, data, ds[vname].attrs)

    for vname_u, vname_v in vector_pairs:
        data_u, data_v = _faces_to_latlon_vector(ds[vname_u].data, ds[vname_v].data,
                                                 nfaces=len(ds['face']))
        data_vars[vname_u] = xr.Variable(_drop_facedim(ds[vname_u].dims), data_u, ds[vname_u].attrs)
        data_vars[vname_v] = xr.Variable(_drop_facedim(ds[vname_v].dims), data_v, ds[vname_v].attrs)
    for vname_u, vname_v in metric_vector_pairs:
        data_u, data_v = _faces_to_latlon_vector(ds[vname_u].data, ds[vname_v].data, nfaces=len(ds['face']), metric=True)
        data_vars[vname_u] = xr.Variable(_drop_facedim(ds[vname_u].dims), data_u, ds[vname_u].attrs)
        data_vars[vname_v] = xr.Variable(_drop_facedim(ds[vname_v].dims), data_v, ds[vname_v].attrs)


    ds_new = ds_new.update(data_vars)
    ds_new = ds_new.set_coords([c for c in coord_vars if c in ds_new])
    return ds_new


# below are data transformers

def _all_facets_to_faces(data_facets, meta, nfaces):
    return {vname: _facets_to_faces(data, nfaces)
            for vname, data in data_facets.items()}


def _all_facets_to_latlon(data_facets, meta, nfaces=None):

    vector_pairs = []
    scalars = []
    vnames = list(data_facets)

    for vname in vnames:
        try:
            mate = meta[vname]['attrs']['mate']
        except KeyError:
            mate = None

        # Raises an exception if the mate of a variable in vnames is missing.
        if mate is not None:
            vector_pairs.append((vname, mate))
            try:
                vnames.remove(mate)
            except ValueError:
                msg = 'If {} in varnames, {} must also be in varnames'.format(vname, mate)
                raise ValueError(msg)

    all_vector_components = [inner for outer in vector_pairs for inner in outer]
    scalars = [vname for vname in vnames if vname not in all_vector_components]

    data = {}
    for vname in scalars:
        data[vname] = _facets_to_latlon_scalar(data_facets[vname])

    for vname_u, vname_v in vector_pairs:
        data_u, data_v = _facets_to_latlon_vector(data_facets[vname_u],
                                                  data_facets[vname_v])
        data[vname_u] = data_u
        data[vname_v] = data_v

    return data

def _chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def _get_facet_chunk(store, varname, iternum, nfacet, klevels, nx, nz, nfaces,
                     dtype, mask_override, domain, pad_before, pad_after):

    fs, path = store.get_fs_and_full_path(varname, iternum)

    assert (nfacet >= 0) & (nfacet < _nfacets)

    file = fs.open(path)

    # insert singleton axis for time (if not grid var) and k level
    facet_shape = (1,) + _facet_shape(nfacet, nx, nfaces)
    facet_shape = (1,) + facet_shape if iternum is not None else facet_shape

    level_data = []


    if (store.shrunk and iternum is not None) or \
       (store.shrunk_grid and iternum is None):
        # the store tells us whether we need a mask or not
        point = _get_variable_point(varname, mask_override)
        mykey = nx if domain == 'global' else f'{domain}_{nx}'
        index = all_index_data[mykey][point]
        zgroup = store.open_mask_group()
        mask = zgroup['mask_' + point].astype('bool')
    else:
        index = None
        mask = None

    # Offset start/end read position due to padding facet before me
    pre_pad = np.cumsum([x+y for x,y in zip(pad_before,pad_after)])
    pre_pad = shift_and_pad(pre_pad,left=False).compute()
    tot_pad = pre_pad[-1]+pad_after[-1]

    for k in klevels:
        assert (k >= 0) & (k < nz)

        # figure out where in the file we have to read to get the data
        # for this level and facet
        if index:
            i = np.ravel_multi_index((k, nfacet), (nz, _nfacets))
            start = index[i]
            end = index[i+1]
        else:
            level_start = k * (nx**2 * nfaces - nx * tot_pad)
            facet_start, facet_end = _uncompressed_facet_index(nfacet, nx, nfaces)
            start = level_start + facet_start
            end = level_start + facet_end - nx*pad_after[nfacet]
            end = end - nx*(pad_before[nfacet]) if k*nfacet==0 else end

            start,end = [x - nx*pre_pad[nfacet] if k+nfacet!=0 else x for x in [start,end]]


        read_offset = start * dtype.itemsize # in bytes
        read_length  = (end - start) * dtype.itemsize # in bytes
        file.seek(read_offset)
        buffer = file.read(read_length)
        data = np.frombuffer(buffer, dtype=dtype)

        assert len(data) == (end - start)

        if mask:
            mask_level = mask[k]
            mask_facets = _faces_to_facets(mask_level,nfaces)
            this_mask = mask_facets[nfacet]
            data = _decompress(data, this_mask, dtype)

        elif pad_before[nfacet]+pad_after[nfacet]>0:

            # Extra care for pad after with rotated fields
            data = _pad_facet(data,facet_shape,_facet_reshape[nfacet],
                              pad_before[nfacet],pad_after[nfacet],dtype)


        # this is the shape this facet is supposed to have
        data.shape = facet_shape
        level_data.append(data)

    out = np.concatenate(level_data, axis=-4)
    return out

def _get_1d_chunk(store, varname, klevels, nz, dtype):
    """for 1D vertical grid variables"""

    fs, path = store.get_fs_and_full_path(varname, None)

    file = fs.open(path)

    # read all levels for 1D variables
    read_length = nz*dtype.itemsize # all levels in bytes
    buffer = file.read(read_length)
    data = np.frombuffer(buffer,dtype=dtype)

    return data[klevels]

class BaseLLCModel:
    """Class representing an LLC Model Dataset.

    Parameters
    ----------
    store : llcreader.BaseStore
        The store object where the data can be found
    mask_ds : zarr.Group
        Must contain variables `mask_c`, `masc_w`, `mask_s`

    Attributes
    ----------
    dtype : numpy.dtype
        Datatype of the data in the dataset
    nx : int
        Number of gridpoints per face (e.g. 90, 1080, 4320, etc.)
    nz : int
        Number of vertical gridpoints
    delta_t : float
        Numerical timestep
    time_units : str
        Date unit string, e.g 'seconds since 1948-01-01 12:00:00'
    iter_start : int
        First model iteration number (inclusive; follows python range conventions)
    iter_stop : int
        Final model iteration number (exclusive; follows python range conventions)
    iter_step : int
        Spacing between iterations
    iters : list of ints
        Specific iteration numbers in a list, possibly with nonuniform spacing.
        Either provide this or the iter parameters above.
    varnames, grid_varnames : list
        List of data variable and grid variable names contained in the dataset
    mask_override : dict
        Override inference of masking variable, e.g. ``{'oceTAUX': 'c'}``
    """

    nface = 13
    dtype = np.dtype('>f4')
    # should be implemented by child classes
    nx = None
    nz = None
    delta_t = None
    time_units = None
    iter_start = None
    iter_stop = None
    iter_step = None
    iters = None
    varnames = []
    grid_varnames = []
    mask_override = {}
    domain = 'global'
    pad_before = [0]*_nfacets
    pad_after  = [0]*_nfacets

    def __init__(self, store):
        """Initialize model

        Parameters
        ----------
        store : llcreader.BaseStore
        mask_ds : zarr.Group
            Must contain variables `mask_c`, `mask_w`, `mask_s`
        """
        self.store = store
        self.shape = (self.nz, self.nface, self.nx, self.nx)
        if self.store.shrunk:
            self.masks = self._get_masks()
            from .shrunk_index import all_index_data
            mykey = self.nx if self.domain == 'global' else f'{self.domain}_{self.nx}'
            self.indexes = all_index_data[mykey]
        else:
            self.masks = None
            self.indexes = None

    def _get_masks(self):
        masks = {}
        zgroup = self.store.open_mask_group()
        for point in ['c', 'w', 's']:
            mask_faces = dsa.from_zarr(zgroup['mask_' + point]).astype('bool')
            masks[point] = _faces_to_facets(mask_faces,self.nface)
        return masks

    def _dtype(self,varname=None):
        if isinstance(self.dtype,np.dtype):
            return self.dtype
        elif isinstance(self.dtype,dict):
            return np.dtype(self.dtype[varname])

    def _get_kp1_levels(self,k_levels):
        # determine kp1 levels
        # get borders to all k (center) levels
        # ki used to get Zu, Zl later
        ku = k_levels[1:] + [k_levels[-1] + 1 ]
        kp1 = []
        ki=[]
        for i,(x,y) in enumerate(zip(k_levels,ku)):
            kp1 += [x] if x not in kp1 else []
            kp1 += [y] if y-x==1 else [x+1]


        kp1=np.array(kp1)

        return kp1

    def _make_coords_faces(self, all_iters):

        time = self.delta_t * all_iters
        time_attrs = {'units': self.time_units,
                      'calendar': self.calendar}
        coords = {'face': ('face', np.arange(self.nface)),
                  'i': ('i', np.arange(self.nx)),
                  'i_g': ('i_g', np.arange(self.nx)),
                  'j': ('j', np.arange(self.nx)),
                  'j_g': ('j_g', np.arange(self.nx)),
                  'k': ('k', np.arange(self.nz)),
                  'k_u': ('k_u', np.arange(self.nz)),
                  'k_l': ('k_l', np.arange(self.nz)),
                  'k_p1': ('k_p1', np.arange(self.nz + 1)),
                  'niter': ('time', all_iters),
                  'time': ('time', time, time_attrs)
                 }
        ds = xr.decode_cf(xr.Dataset(coords=coords))

        from ..variables import dimensions
        for d in dimensions:
            if d in ds:
                ds[d].attrs.update(dimensions[d]['attrs'])
        return ds

    def _make_coords_latlon():
        ds = self._make_coords_faces(self)
        return _faces_coords_to_latlon(ds)


    def _dask_array(self, nfacet, varname, iters, klevels, k_chunksize):
        # return a dask array for a single facet
        facet_shape =  _facet_shape(nfacet, self.nx, self.nface)
        time_chunks = (len(iters) * (1,),) if iters is not None else ()
        k_chunks = (tuple([len(c)
                          for c in _chunks(klevels, k_chunksize)]),)
        chunks = time_chunks + k_chunks + tuple([(s,) for s in facet_shape])

        # manually build dask graph
        dsk = {}
        token = tokenize(varname, self.store, nfacet)
        name = '-'.join([varname, token])
        dtype = self._dtype(varname)

        # iters == None for grid variables
        def _key_and_task(n_k, these_klevels, n_iter=None, iternum=None):
            if n_iter is None:
                key = name, n_k, 0, 0, 0
            else:
                key = name, n_iter, n_k, 0, 0, 0
            task = (_get_facet_chunk, self.store, varname, iternum,
                     nfacet, these_klevels, self.nx, self.nz, self.nface,
                     dtype, self.mask_override, self.domain,
                     self.pad_before, self.pad_after)
            return key, task

        if iters is not None:
            for n_iter, iternum in enumerate(iters):

                for n_k, these_klevels in enumerate(_chunks(klevels, k_chunksize)):
                    key, task = _key_and_task(n_k, these_klevels, n_iter, iternum)
                    dsk[key] = task
        else:
            for n_k, these_klevels in enumerate(_chunks(klevels, k_chunksize)):
                key, task = _key_and_task(n_k, these_klevels)
                dsk[key] = task

        return dsa.Array(dsk, name, chunks, dtype)

    def _dask_array_vgrid(self, varname, klevels, k_chunksize):
        # return a dask array for a 1D vertical grid var

        # single chunk for 1D variables
        chunks = ((len(klevels),),)

        # manually build dask graph
        dsk = {}
        token = tokenize(varname, self.store)
        name = '-'.join([varname, token])
        dtype = self._dtype(varname)

        nz = self.nz if _VAR_METADATA[varname]['dims'] != ['k_p1'] else self.nz+1
        task = (_get_1d_chunk, self.store, varname,
                list(klevels), nz, dtype)

        key = name, 0
        dsk[key] = task

        return dsa.Array(dsk, name, chunks, dtype)

    def _get_facet_data(self, varname, iters, klevels, k_chunksize):
        # needs facets to be outer index of nested lists
        dims = _VAR_METADATA[varname]['dims']

        if len(dims)==2:
            klevels = [0,]

        if _is_vgrid(varname):
            data_facets = self._dask_array_vgrid(varname,klevels,k_chunksize)
        else:
            data_facets = [self._dask_array(nfacet, varname, iters, klevels, k_chunksize)
                       for nfacet in range(5)]

        if len(dims)==2:
            # squeeze depth dimension out of 2D variable
            data_facets = [facet[..., 0, :, :, :] for facet in data_facets]

        return data_facets


    def _check_iter_start(self, iter_start):
        if self.iter_start is not None and self.iter_step is not None:
            if (iter_start - self.iter_start) % self.iter_step:
                msg = "Iteration {} may not exist, you may need to change 'iter_start'".format(iter_start)
                warnings.warn(msg, RuntimeWarning)


    def _check_iter_step(self, iter_step):
        if self.iter_step is not None:
            if iter_step % self.iter_step:
                msg = "'iter_step' is not a multiple of {}, meaning some expected timesteps may not be returned".format(self.iter_step)
                warnings.warn(msg, RuntimeWarning)

    def _check_iters(self, iters):
        if self.iters is not None:
            if not set(iters) <= set(self.iters):
                msg = "Some requested iterations may not exist, you may need to change 'iters'"
                warnings.warn(msg, RuntimeWarning)
        
        elif self.iter_start is not None and self.iter_step is not None:
            for iter in iters:
                if (iter - self.iter_start) % self.iter_step:
                    msg = "Some requested iterations may not exist, you may need to change 'iters'"
                    warnings.warn(msg, RuntimeWarning)
                    break
            

    def get_dataset(self, varnames=None, iter_start=None, iter_stop=None,
                    iter_step=None, iters=None, k_levels=None, k_chunksize=1,
                    type='faces', read_grid=True, grid_vars_to_coords=True):
        """
        Create an xarray Dataset object for this model.

        Parameters
        ----------
        *varnames : list of strings, optional
            The variables to include, e.g. ``['Salt', 'Theta']``. Otherwise
            include all known variables.
        iter_start : int, optional
            Starting iteration number. Otherwise use model default.
            Follows standard `range` conventions. (inclusive)
        iter_stop : int, optional
            Stopping iteration number. Otherwise use model default.
            Follows standard `range` conventions. (exclusive)
        iter_step : int, optional
            Iteration number stepsize. Otherwise use model default.
        iters : list of ints, optional
            Specific iteration numbers in a list, possibly with nonuniform spacing.
            Either provide this or the iter parameters above.
        k_levels : list of ints, optional
            Vertical levels to extract. Default is to get them all
        k_chunksize : int, optional
            How many vertical levels per Dask chunk.
        type : {'faces', 'latlon'}, optional
            What type of dataset to create
        read_grid : bool, optional
            Whether to read the grid info
        grid_vars_to_coords : bool, optional
            Whether to promote grid variables to coordinate status

        Returns
        -------
        ds : xarray.Dataset
        """

        def _if_not_none(a, b):
            if a is None:
                return b
            else:
                return a

        user_iter_params = [iter_start, iter_stop, iter_step]
        attribute_iter_params = [self.iter_start, self.iter_stop, self.iter_step]

        # If the user has specified some iter params:
        if any([a is not None for a in user_iter_params]):
            # If iters is also set we have a problem
            if iters is not None:
                raise ValueError("Only `iters` or the parameters `iter_start`, `iters_stop`, "
                                 "and `iter_step` can be provided. Both were provided")
            
            # Otherwise we can override any missing values
            iter_start = _if_not_none(iter_start, self.iter_start)
            iter_stop = _if_not_none(iter_stop, self.iter_stop)
            iter_step = _if_not_none(iter_step, self.iter_step)
            iter_params = [iter_start, iter_stop, iter_step]
            if any([a is None for a in iter_params]):
                raise ValueError("The parameters `iter_start`, `iter_stop`, "
                                 "and `iter_step` must be defined either by the "
                                 "model class or as argument. Instead got %r "
                                 % iter_params)
        
        # Otherwise try loading from the user set iters
        elif iters is not None:
            pass

        # Now have a go at using the attribute derived iteration parameters 
        elif all([a is not None for a in attribute_iter_params]):
            iter_params = attribute_iter_params

        # Now try using the attribute derived iters
        elif self.iters is not None:
            iters = self.iters

        # Now give up
        else:
            raise ValueError("The parameters `iter_start`, `iter_stop`, "
                             "and `iter_step`, or `iters` must be defined either by the "
                             "model class or as argument")
        
        # Check the iter_start and iter_step
        if iters is None:
            self._check_iter_start(iter_params[0])
            self._check_iter_step(iter_params[2])
            iters = np.arange(*iter_params)
        else:
            self._check_iters(iters)
            iters = np.array(iters)

        varnames = varnames or self.varnames

        # grid stuff
        read_grid = read_grid and len(self.grid_varnames)!=0
        grid_vars_to_coords = grid_vars_to_coords and read_grid
        grid_varnames = self.grid_varnames if read_grid else []

        ds = self._make_coords_faces(iters)
        if type=='latlon':
            if self.domain=='aste':
                raise TypeError('Swapping to lat/lon not available for ASTE. Must regrid or interpolate.')
            ds = _faces_coords_to_latlon(ds)

        k_levels = k_levels or list(range(self.nz))
        kp1_levels = self._get_kp1_levels(k_levels)

        ds = ds.sel(k=k_levels, k_l=k_levels, k_u=k_levels, k_p1=kp1_levels)

        # get the data in facet form
        data_facets = {vname:
                       self._get_facet_data(vname, iters, k_levels, k_chunksize)
                       for vname in varnames}

        # get the grid in facet form
        # do separately for vertical coords on kp1_levels
        grid_facets = {}
        for vname in grid_varnames:
            my_k_levels = k_levels if _VAR_METADATA[vname]['dims'] !=['k_p1'] else kp1_levels
            grid_facets[vname] = self._get_facet_data(vname, None, my_k_levels, k_chunksize)

        # transform it into faces or latlon
        data_transformers = {'faces': _all_facets_to_faces,
                             'latlon': _all_facets_to_latlon}

        transformer = data_transformers[type]
        data = transformer(data_facets, _VAR_METADATA, self.nface)

        # separate horizontal and vertical grid variables
        hgrid_facets = {key: grid_facets[key]
                for key in grid_varnames if not _is_vgrid(key)}
        vgrid_facets = {key: grid_facets[key]
                for key in grid_varnames if _is_vgrid(key)}

        # do not transform vertical grid variables
        data.update(transformer(hgrid_facets, _VAR_METADATA, self.nface))
        data.update(vgrid_facets)

        variables = {}
        gridlist = ['Zl','Zu'] if read_grid else []
        for vname in varnames+grid_varnames:
            meta = _VAR_METADATA[vname]
            dims = meta['dims']
            if type=='faces':
                dims = _add_face_to_dims(dims)
            dims = ['time',] + dims if vname not in grid_varnames else dims
            attrs = meta['attrs']

            # Handle grid names different from filenames
            fname = vname
            vname = meta['real_name'] if 'real_name' in meta else vname
            if fname in grid_varnames:
                gridlist.append(vname)

            variables[vname] = xr.Variable(dims, data[fname], attrs)

        # handle vertical coordinate after the fact
        if read_grid and 'RF' in grid_varnames:
            ki = np.array([list(kp1_levels).index(x) for x in k_levels])
            for zv,sl in zip(['Zl','Zu'],[ki,ki+1]):
                variables[zv] = xr.Variable(_VAR_METADATA[zv]['dims'],
                                            data['RF'][sl],
                                            _VAR_METADATA[zv]['attrs'])

        ds = ds.update(variables)

        if grid_vars_to_coords:
            ds = ds.set_coords(gridlist)

        return ds



import fsspec
import os
import zarr


class BaseStore:
    """Basic storage class for LLC data.

    Parameters
    ----------
    fs : fsspec.AbstractFileSystem
    base_path : str, optional
        Where to find the data within the filesystem
    shrunk : bool, optional
        Whether the data files have been tagged with `.shrunk`
    mask_fs, grid_fs : fsspec.AbstractFileSystem, optional
        Where to find the mask or grid datasets to decode the compression
    mask_path, grid_path : str, optional
        Path to the mask or grid datasets on the ``mask_fs`` or ``grid_fs`` filesystem
    shrunk_grid : bool, optional
        Whether the grid files have been tagged with `.shrunk`
        not always the same as for data variables
    join_char : str or None
        Character to use to join paths. Falls back on os.path.join if None.
    """

    def __init__(self, fs, base_path='/', shrunk=False,
                 mask_fs=None, mask_path=None, 
                 grid_fs=None, grid_path=None,
                 shrunk_grid=False, join_char=None):
        self.base_path = base_path
        self.fs = fs
        self.shrunk = shrunk
        self.mask_fs = mask_fs or self.fs
        self.mask_path = mask_path
        self.grid_fs = grid_fs or self.fs
        self.grid_path = grid_path
        self.shrunk_grid = shrunk_grid
        self.join_char = join_char
        if shrunk and (mask_path is None):
            raise ValueError("`mask_path` can't be None if `shrunk` is True")


    def _directory(self, varname, iternum):
        if iternum is not None:
            return self.base_path
        else:
            return self.grid_path

    def _fname(self, varname, iternum):

        if iternum is not None:
            fname = varname + '.%010d.data' % iternum
            if self.shrunk:
                fname += '.shrunk'
        else:
            fname = varname + '.data'
            if self.shrunk_grid:
                fname = varname + '.shrunk'

        return fname

    def _join(self, *args):
        if self.join_char:
            return self.join_char.join(args)
        else:
            return os.path.join(*args)

    def _full_path(self, varname, iternum):
        return self._join(self._directory(varname, iternum),
                            self._fname(varname, iternum))

    def get_fs_and_full_path(self, varname, iternum):
        """Return references to a filesystem and path within it for a specific
        variable and iteration number.

        Parameters
        ----------
        varname : str
        iternum : int

        Returns
        -------
        fs : fsspec.AbstractFileSystem
            The filesytem where the file can be found
        path : str
            The path to open
        """
        return self.fs, self._full_path(varname, iternum)

    def open_data_file(self, varname, iternum):
        """Open the file for a specific variable and iteration number.

        Parameters
        ----------
        varname : str
        iternum : int

        Returns
        -------
        fobj : file-like object
        """
        fs, path = self.get_fs_and_full_path(varname, iternum)
        return fs.open(path)

    def open_mask_group(self):
        """Open the zarr group that contains the masks

        Returns
        -------
        mask_group : zarr.Group
        """

        mapper = self.mask_fs.get_mapper(self.mask_path)
        zgroup = zarr.open_consolidated(mapper)
        return zgroup


class NestedStore(BaseStore):
    """Store where the variable are stored in subdirectories according to
    iteration number."""

    def _directory(self, varname, iternum):
        if iternum is not None:
            return self._join(self.base_path, '%010d' % iternum)
        else:
            return self.grid_path
